"""
The :py:mod:`vc2_conformance.bitstream.vc2_autofill` module provides auto-fill
routines for automatically computing certain values for the context
dictionaries used by the :py:mod:`vc2_conformance.bitstream.vc2`
:py:mod:`~vc2_conformance.bitstream.serdes` functions. These values include the
picture number and parse offset fields which can't default to a simple fixed
value.

In the common case, the :py:func:`autofill_and_serialise_stream` function may
be used to serialise a complete :py:class:`~vc2_conformance.bitstream.Stream`,
with sensible defaults provided for all fields (including picture numbers and
next/previous parse offsets).

.. autofunction:: autofill_and_serialise_stream

Autofill value routines
-----------------------

The following functions implement autofill routines for specific bitstream
values.

.. autofunction:: autofill_picture_number

.. autofunction:: autofill_major_version

.. autofunction:: autofill_parse_offsets

.. autofunction:: autofill_parse_offsets_finalize


Autofill value dictionary
-------------------------

.. autodata:: vc2_default_values_with_auto
    :annotation:

.. autodata:: AUTO
    :annotation:

"""

from copy import deepcopy

from sentinels import Sentinel

from vc2_data_tables import (
    PARSE_INFO_HEADER_BYTES,
    ParseCodes,
)

from vc2_conformance.version_constraints import (
    MINIMUM_MAJOR_VERSION,
    preset_frame_rate_version_implication,
    preset_signal_range_version_implication,
    preset_color_spec_version_implication,
    preset_color_primaries_version_implication,
    preset_color_matrix_version_implication,
    preset_transfer_function_version_implication,
    wavelet_transform_version_implication,
    parse_code_version_implication,
    profile_version_implication,
)

from vc2_conformance.pseudocode.state import State

from vc2_conformance.bitstream.io import BitstreamWriter
from vc2_conformance.bitstream.serdes import Serialiser

from vc2_conformance.bitstream.vc2 import parse_stream

from vc2_conformance.bitstream.vc2_fixeddicts import (
    vc2_default_values,
    ParseInfo,
    AuxiliaryData,
    Padding,
    ParseParameters,
    PictureParse,
    PictureHeader,
    FragmentParse,
    FragmentHeader,
    FrameRate,
    SignalRange,
    ColorSpec,
    ColorPrimaries,
    ColorMatrix,
    TransferFunction,
    TransformParameters,
    ExtendedTransformParameters,
)

from vc2_conformance.pseudocode.parse_code_functions import (
    is_picture,
    is_fragment,
)


__all__ = [
    "AUTO",
    "vc2_default_values_with_auto",
    "autofill_picture_number",
    "autofill_major_version",
    "autofill_parse_offsets",
    "autofill_parse_offsets_finalize",
    "autofill_and_serialise_stream",
]


AUTO = Sentinel("AUTO")
"""
A constant which may be placed in a
:py:mod:`~vc2_conformance.bitstream.vc2_fixeddicts` fixed dictionary field to
indicate that the various ``autofill_*`` functions in this module should
automatically compute a value for that field.
"""

vc2_default_values_with_auto = deepcopy(vc2_default_values)
"""
Like :py:data:`vc2_conformance.bitstreams.vc2_default_values` but with
:py:data:`AUTO` set as the default value for all fields which support it.
"""

vc2_default_values_with_auto[ParseInfo]["next_parse_offset"] = AUTO
vc2_default_values_with_auto[ParseInfo]["previous_parse_offset"] = AUTO
vc2_default_values_with_auto[ParseParameters]["major_version"] = AUTO
vc2_default_values_with_auto[PictureHeader]["picture_number"] = AUTO
vc2_default_values_with_auto[FragmentHeader]["picture_number"] = AUTO


def get_auto(d, field_name, dtype):
    """
    For internal use. Get a value from the dictionary d, falling back on the
    :py:data:`vc2_default_values_with_auto` default value.

    Parameters
    ==========
    d : dict
    field_name: str
    dtype: :py:class:~`vc2_conformance.bitstream.vc2_fixeddicts` instance
    """
    if field_name in d:
        return d[field_name]
    else:
        return vc2_default_values_with_auto[dtype][field_name]


def get_transform_parameters(data_unit):
    """
    For internal use. Given a
    :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.DataUnit`, if the data
    unit contains a picture or the first fragment of a fragmented picture,
    returns the
    :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.TransformParameters`,
    creating an empty one if it is not defined.  Otherwise returns None.
    """
    parse_code = get_auto(data_unit.get("parse_info", {}), "parse_code", ParseInfo)
    if is_picture(State(parse_code=parse_code)):
        return (
            data_unit.setdefault("picture_parse", {})
            .setdefault("wavelet_transform", {})
            .setdefault("transform_parameters", {})
        )
    elif (
        is_fragment(State(parse_code=parse_code))
        and get_auto(
            data_unit.get("fragment_parse", {}).get("fragment_header", {}),
            "fragment_slice_count",
            FragmentHeader,
        )
        == 0
    ):
        return data_unit.setdefault("fragment_parse", {}).setdefault(
            "transform_parameters", {}
        )
    else:
        return None


def autofill_picture_number(stream, initial_picture_number=0):
    """
    Given a :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.Stream`,
    find all picture_number fields which are absent or contain the
    :py:data:`AUTO` sentinel and automatically fill them with consecutive
    picture numbers. Numbering is restarted for each sequence.
    """
    for sequence in stream.get("sequences", []):
        last_picture_number = (initial_picture_number - 1) & 0xFFFFFFFF

        for data_unit in sequence.get("data_units", []):
            parse_code = data_unit.get("parse_info", {}).get("parse_code")

            # Get the current picture/fragment header (in 'header') and determine if
            # the picture number should be incremented in this picture/fragment or
            # not ('increment' is True if an incremented picture number should be
            # used, False otherwise)
            if parse_code in (
                ParseCodes.low_delay_picture,
                ParseCodes.high_quality_picture,
            ):
                picture_parse = data_unit.setdefault("picture_parse", PictureParse())
                header = picture_parse.setdefault("picture_header", PictureHeader())
                increment = True
            elif parse_code in (
                ParseCodes.low_delay_picture_fragment,
                ParseCodes.high_quality_picture_fragment,
            ):
                fragment_parse = data_unit.setdefault("fragment_parse", FragmentParse())
                header = fragment_parse.setdefault("fragment_header", FragmentHeader())
                increment = (
                    header.get(
                        "fragment_slice_count",
                        vc2_default_values_with_auto[FragmentHeader][
                            "fragment_slice_count"
                        ],
                    )
                    == 0
                )
            else:
                # Not a picture; move on!
                continue

            if header.get("picture_number", AUTO) is AUTO:
                if increment:
                    header["picture_number"] = (last_picture_number + 1) & 0xFFFFFFFF
                else:
                    header["picture_number"] = last_picture_number

            last_picture_number = header["picture_number"]


def autofill_major_version(stream):
    """
    Given a :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.Stream`, find
    all ``major_version`` fields which are set to the :py:data:`AUTO` sentinel
    and automatically set them to the appropriate version number for the
    features used by this stream.

    As a side effect, this function will automatically remove the
    :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.ExtendedTransformParameters`
    field whenever it appears in
    :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.TransformParameters`
    if the major_version evaluates to less than 3. This change will only be
    made when ``major_version`` was set to AUTO in a proceeding sequence
    header, if the field was explicitly set to a particular value, no changes
    will be made to any transform parameters dicts which follow.
    """
    for sequence in stream.get("sequences", []):
        # Compute the major version number to be used according to (11.2.2)
        major_version = MINIMUM_MAJOR_VERSION
        for data_unit in sequence.get("data_units", []):
            parse_code = get_auto(
                data_unit.get("parse_info", {}), "parse_code", ParseInfo
            )

            # Check parse code version requirements
            major_version = max(
                major_version, parse_code_version_implication(parse_code)
            )

            if parse_code == ParseCodes.sequence_header:
                sequence_header = data_unit.get("sequence_header", {})

                # Check profile version requirements
                parse_parameters = sequence_header.get("parse_parameters", {})
                profile = get_auto(parse_parameters, "profile", ParseParameters)
                major_version = max(major_version, profile_version_implication(profile))

                # Check video parameter preset version parameters

                source_parameters = sequence_header.get("video_parameters", {})

                frame_rate = source_parameters.get("frame_rate", {})
                if get_auto(frame_rate, "custom_frame_rate_flag", FrameRate):
                    index = get_auto(frame_rate, "index", FrameRate)
                    major_version = max(
                        major_version, preset_frame_rate_version_implication(index)
                    )

                signal_range = source_parameters.get("signal_range", {})
                if get_auto(signal_range, "custom_signal_range_flag", SignalRange):
                    index = get_auto(signal_range, "index", SignalRange)
                    major_version = max(
                        major_version, preset_signal_range_version_implication(index)
                    )

                color_spec = source_parameters.get("color_spec", {})
                if get_auto(color_spec, "custom_color_spec_flag", ColorSpec):
                    index = get_auto(color_spec, "index", ColorSpec)
                    major_version = max(
                        major_version, preset_color_spec_version_implication(index)
                    )

                    if index == 0:
                        color_primaries = color_spec.get("color_primaries", {})
                        if get_auto(
                            color_primaries,
                            "custom_color_primaries_flag",
                            ColorPrimaries,
                        ):
                            index = get_auto(color_primaries, "index", ColorPrimaries)
                            major_version = max(
                                major_version,
                                preset_color_primaries_version_implication(index),
                            )

                        color_matrix = color_spec.get("color_matrix", {})
                        if get_auto(
                            color_matrix, "custom_color_matrix_flag", ColorMatrix
                        ):
                            index = get_auto(color_matrix, "index", ColorMatrix)
                            major_version = max(
                                major_version,
                                preset_color_matrix_version_implication(index),
                            )

                        transfer_function = color_spec.get("transfer_function", {})
                        if get_auto(
                            transfer_function,
                            "custom_transfer_function_flag",
                            TransferFunction,
                        ):
                            index = get_auto(
                                transfer_function, "index", TransferFunction
                            )
                            major_version = max(
                                major_version,
                                preset_transfer_function_version_implication(index),
                            )
            else:
                # Check wavelet symmetry version parameters
                tp = get_transform_parameters(data_unit)
                if tp is not None:
                    etp = tp.get("extended_transform_parameters", {})

                    wavelet_index = get_auto(tp, "wavelet_index", TransformParameters)

                    wavelet_index_ho = wavelet_index
                    if get_auto(
                        etp, "asym_transform_index_flag", ExtendedTransformParameters
                    ):
                        wavelet_index_ho = get_auto(
                            etp, "wavelet_index_ho", ExtendedTransformParameters
                        )

                    dwt_depth_ho = 0
                    if get_auto(
                        etp, "asym_transform_flag", ExtendedTransformParameters
                    ):
                        dwt_depth_ho = get_auto(
                            etp, "dwt_depth_ho", ExtendedTransformParameters
                        )

                    major_version = max(
                        major_version,
                        wavelet_transform_version_implication(
                            wavelet_index, wavelet_index_ho, dwt_depth_ho
                        ),
                    )

        # Modify sequence headers to include the correct version number,
        # additionally, remove ExtendedTransformParameters where defined when
        # AUTO has been used. Where AUTO has been used,
        # ExtendedTransformParameters can be safely removed when major_version
        # is not 3 (since a symmetric transform must have been applied). When
        # AUTO has not been used, we leave it up to the user to make the
        # version and ETP existance match up.
        auto_used = False
        for data_unit in sequence.get("data_units", []):
            parse_code = get_auto(
                data_unit.get("parse_info", {}), "parse_code", ParseInfo
            )
            if parse_code == ParseCodes.sequence_header:
                # Case: Auto-set version number if required
                sequence_header = data_unit.setdefault("sequence_header", {})
                parse_parameters = sequence_header.setdefault("parse_parameters", {})
                if get_auto(parse_parameters, "major_version", ParseParameters) is AUTO:
                    parse_parameters["major_version"] = major_version
                    auto_used = True
                else:
                    auto_used = False
            else:
                tp = get_transform_parameters(data_unit)
                if tp is not None and auto_used:
                    # Case: Remove ExtendedTransformParameters if required
                    if major_version < 3 and "extended_transform_parameters" in tp:
                        del tp["extended_transform_parameters"]


def autofill_parse_offsets(stream):
    """
    Given a :py:class:`~vc2_conformance.bitstream.Stream`,
    find and fill in all next_parse_offset and previous_parse_offset fields
    which are absent or contain the :py:data:`AUTO` sentinel.

    In many (but not all) cases computing these field values is most
    straight-forwardly done post serialisation. In these cases, fields in the
    stream will be autofilled with '0'. These fields should then subsequently
    be ammended by :py:func:`autofill_parse_offsets_finalize`.
    """
    next_parse_offsets_to_autofill = []
    previous_parse_offsets_to_autofill = []

    for sequence_index, sequence in enumerate(stream.get("sequences", [])):
        for data_unit_index, data_unit in enumerate(sequence.get("data_units", [])):
            parse_info = data_unit.setdefault("parse_info", ParseInfo())
            parse_code = parse_info.get("parse_code")

            if parse_code in (ParseCodes.auxiliary_data, ParseCodes.padding_data):
                # The length of padding and aux. data fields are determined by the
                # next_parse_offset field so these should be auto-fillled based on the
                # length of padding/aux data present.
                if parse_info.get("next_parse_offset", AUTO) is AUTO:
                    if parse_code == ParseCodes.auxiliary_data:
                        data = data_unit.get("auxiliary_data", {}).get(
                            "bytes",
                            vc2_default_values_with_auto[AuxiliaryData]["bytes"],
                        )
                    elif parse_code == ParseCodes.padding_data:
                        data = data_unit.get("padding", {}).get(
                            "bytes", vc2_default_values_with_auto[Padding]["bytes"]
                        )
                    parse_info["next_parse_offset"] = PARSE_INFO_HEADER_BYTES + len(
                        data
                    )

            if parse_info.get("next_parse_offset", AUTO) is AUTO:
                parse_info["next_parse_offset"] = 0
                next_parse_offsets_to_autofill.append((sequence_index, data_unit_index))

            if parse_info.get("previous_parse_offset", AUTO) is AUTO:
                parse_info["previous_parse_offset"] = 0
                previous_parse_offsets_to_autofill.append(
                    (sequence_index, data_unit_index)
                )

    return (next_parse_offsets_to_autofill, previous_parse_offsets_to_autofill)


def autofill_parse_offsets_finalize(
    bitstream_writer,
    stream,
    next_parse_offsets_to_autofill,
    previous_parse_offsets_to_autofill,
):
    """
    Finalize the autofillling of next and previous parse offsets by directly
    modifying the serialised bitstream.

    Parameters
    ==========
    bitstream_writer : :py:class:`~vc2_conformance.bitstream.io.BitstreamWriter`
        A :py:class:`~vc2_conformance.bitstream.io.BitstreamWriter` set up to
        write to the already-serialised bitstream.
    stream : :py:class:`~vc2_conformance.bitstream.Stream`
        The context dictionary used to serialies the bitstream. Since computed
        values added to these dictionaries by the serialisation process, it may
        be necessary to use the dictionary provided by
        :py:attr:`vc2_conformance.bitstream.Serialiser.context`, rather than
        the one passed into the Serialiser. This is because the Serialiser may
        have replaced some dictionaries during serialisation.
    next_parse_offsets_to_autofill, previous_parse_offsets_to_autofill
        The arrays of parse info indices whose next and previous parse offsets
        remain to be auto-filled.
    """
    end_of_sequence_offset = bitstream_writer.tell()

    for sequence_index, data_unit_index in next_parse_offsets_to_autofill:
        sequence = stream["sequences"][sequence_index]

        if data_unit_index == len(sequence["data_units"]) - 1:
            next_parse_offset = 0
        else:
            next_parse_offset = (
                sequence["data_units"][data_unit_index + 1]["parse_info"]["_offset"]
                - sequence["data_units"][data_unit_index]["parse_info"]["_offset"]
            )
        byte_offset = sequence["data_units"][data_unit_index]["parse_info"]["_offset"]
        bitstream_writer.seek(byte_offset + 4 + 1)  # Seek past prefix and parse code
        bitstream_writer.write_uint_lit(4, next_parse_offset)
        bitstream_writer.flush()

    for sequence_index, data_unit_index in previous_parse_offsets_to_autofill:
        sequence = stream["sequences"][sequence_index]

        if data_unit_index == 0:
            previous_parse_offset = 0
        else:
            previous_parse_offset = (
                sequence["data_units"][data_unit_index]["parse_info"]["_offset"]
                - sequence["data_units"][data_unit_index - 1]["parse_info"]["_offset"]
            )
        byte_offset = sequence["data_units"][data_unit_index]["parse_info"]["_offset"]
        bitstream_writer.seek(
            byte_offset + 4 + 1 + 4
        )  # Seek past prefix, parse code and next offset
        bitstream_writer.write_uint_lit(4, previous_parse_offset)
        bitstream_writer.flush()

    bitstream_writer.seek(*end_of_sequence_offset)


def autofill_and_serialise_stream(file, stream):
    """
    Given a :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.Stream`
    dictionary describing a VC-2 stream, serialise that into the supplied
    file.

    Parameters
    ==========
    file : file-like object
        A file open for binary writing. The serialised bitstream will be
        written to this file.
    stream : :py:class:`~vc2_conformance.Stream`
        The stream to be serialised. Unspecified values will be auto-filled if
        possible. See :ref:`bitstream-fixeddicts` for the default auto-fill
        values.

        .. note::

            Internally, auto-fill values are taken from
            :py:data:`vc2_default_values_with_auto`.

            Supported fields containing the special value :py:class:`AUTO` will
            be autofilled with suitably computed values. Specifically:

            * Picture numbers will set to incrementing values (starting at 0,
              or continuing from the value used by the previous picture) by
              :py:func:`autofill_picture_number`.
            * The ``major_version`` field will be populated by
              :py:func:`autofill_major_version` and, if appropriate, extended
              transform parameters fields will be removed.
            * Next and previous parse offsets will be calculated automatically
              by
              :py:func:`autofill_parse_offsets`
              and
              :py:func:`autofill_parse_offsets_finalize`.
    """
    autofill_picture_number(stream)
    autofill_major_version(stream)
    (
        next_parse_offsets_to_autofill,
        previous_parse_offsets_to_autofill,
    ) = autofill_parse_offsets(stream)

    writer = BitstreamWriter(file)
    with Serialiser(writer, stream, vc2_default_values_with_auto) as serdes:
        parse_stream(serdes, State())
    writer.flush()

    autofill_parse_offsets_finalize(
        writer,
        serdes.context,
        next_parse_offsets_to_autofill,
        previous_parse_offsets_to_autofill,
    )



"""
The :py:mod:`vc2_conformance.decoder.sequence_header` module contains pseudocode
functions from (11) Sequence Header.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_data_tables import (
    BaseVideoFormats,
    PictureCodingModes,
    Profiles,
    Levels,
    ColorDifferenceSamplingFormats,
    SourceSamplingModes,
    PresetFrameRates,
    PresetPixelAspectRatios,
    PresetSignalRanges,
    PresetColorSpecs,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
)

from vc2_conformance.level_constraints import LEVEL_SEQUENCE_RESTRICTIONS

from vc2_conformance.version_constraints import (
    MINIMUM_MAJOR_VERSION,
    preset_frame_rate_version_implication,
    preset_signal_range_version_implication,
    preset_color_spec_version_implication,
    preset_color_primaries_version_implication,
    preset_color_matrix_version_implication,
    preset_transfer_function_version_implication,
    profile_version_implication,
)

from vc2_conformance.pseudocode.video_parameters import (
    set_source_defaults,
    set_coding_parameters,
    preset_frame_rate,
    preset_pixel_aspect_ratio,
    preset_signal_range,
    preset_color_primaries,
    preset_color_matrix,
    preset_transfer_function,
    preset_color_spec,
)

from vc2_conformance.decoder.assertions import (
    assert_in_enum,
    assert_level_constraint,
    log_version_lower_bound,
)

from vc2_conformance.decoder.exceptions import (
    SequenceHeaderChangedMidSequence,
    BadBaseVideoFormat,
    BadPictureCodingMode,
    BadProfile,
    BadLevel,
    BadColorDifferenceSamplingFormat,
    BadSourceSamplingMode,
    BadPresetFrameRateIndex,
    BadPresetPixelAspectRatio,
    CleanAreaOutOfRange,
    BadCustomSignalExcursion,
    BadPresetSignalRange,
    BadPresetColorSpec,
    BadPresetColorPrimaries,
    BadPresetColorMatrix,
    BadPresetTransferFunction,
    PictureDimensionsNotMultipleOfFrameDimensions,
    ZeroPixelFrameSize,
    FrameRateHasZeroNumerator,
    FrameRateHasZeroDenominator,
    PixelAspectRatioContainsZeros,
    PresetFrameRateNotSupportedByVersion,
    PresetSignalRangeNotSupportedByVersion,
    PresetColorSpecNotSupportedByVersion,
    PresetColorPrimariesNotSupportedByVersion,
    PresetColorMatrixNotSupportedByVersion,
    PresetTransferFunctionNotSupportedByVersion,
    ProfileNotSupportedByVersion,
    MinorVersionNotZero,
    MajorVersionTooLow,
)

from vc2_conformance.symbol_re import Matcher

from vc2_conformance.decoder.io import (
    record_bitstream_start,
    record_bitstream_finish,
    tell,
    read_bool,
    read_uint,
)


__all__ = [
    "sequence_header",
    "parse_parameters",
    "source_parameters",
    "frame_size",
    "color_diff_sampling_format",
    "scan_format",
    "frame_rate",
    "pixel_aspect_ratio",
    "clean_area",
    "signal_range",
    "color_spec",
    "color_primaries",
    "color_matrix",
    "transfer_function",
]


@ref_pseudocode
def sequence_header(state):
    """(11.1)"""
    # Record this sequence_header as it appears in the bitstream
    ## Begin not in spec
    this_sequence_header_offset = tell(state)[0]
    record_bitstream_start(state)
    ## End not in spec

    parse_parameters(state)

    base_video_format = read_uint(state)

    # (11.4.1) Check base video format is supported
    ## Begin not in spec
    assert_in_enum(base_video_format, BaseVideoFormats, BadBaseVideoFormat)
    ## End not in spec

    # (C.3) Check level allows this base format
    ## Begin not in spec
    assert_level_constraint(state, "base_video_format", base_video_format)
    ## End not in spec

    video_parameters = source_parameters(state, base_video_format)

    state["picture_coding_mode"] = read_uint(state)

    # (11.5) Ensure picture coding mode is valid
    ## Begin not in spec
    assert_in_enum(
        state["picture_coding_mode"], PictureCodingModes, BadPictureCodingMode
    )
    ## End not in spec

    # (C.3) Check level allows this picture coding mode
    ## Begin not in spec
    assert_level_constraint(state, "picture_coding_mode", state["picture_coding_mode"])
    ## End not in spec

    set_coding_parameters(state, video_parameters)

    # Errata: The spec only says the frame_height must be an integer multiple
    # of the color_diff_height but, in addition, the luma_height should be a
    # multiple and the color_diff_width should be a multiple of
    # luma_width/frame_width
    #
    # (11.6.2) Check frame_height is an integer multiple of color_diff_height
    ## Begin not in spec
    if (
        # Prevent divide-by-zero in tests below
        state["luma_height"] == 0
        or state["luma_width"] == 0
        or state["color_diff_height"] == 0
        or state["color_diff_width"] == 0
        or
        # Actually check multiples
        video_parameters["frame_height"] % state["luma_height"] != 0
        or video_parameters["frame_width"] % state["luma_width"] != 0
        or video_parameters["frame_height"] % state["color_diff_height"] != 0
        or video_parameters["frame_width"] % state["color_diff_width"] != 0
    ):
        raise PictureDimensionsNotMultipleOfFrameDimensions(
            state["luma_width"],
            state["luma_height"],
            state["color_diff_width"],
            state["color_diff_height"],
            video_parameters["frame_width"],
            video_parameters["frame_height"],
        )
    ## End not in spec

    # (11.1) Check that the this sequence_header is byte-for-byte identical
    # with the previous sequence_header in the sequence
    ## Begin not in spec
    this_sequence_header_bytes = record_bitstream_finish(state)
    if "_last_sequence_header_bytes" in state:
        if this_sequence_header_bytes != state["_last_sequence_header_bytes"]:
            raise SequenceHeaderChangedMidSequence(
                state["_last_sequence_header_offset"],
                state["_last_sequence_header_bytes"],
                this_sequence_header_offset,
                this_sequence_header_bytes,
            )
    state["_last_sequence_header_bytes"] = this_sequence_header_bytes
    state["_last_sequence_header_offset"] = this_sequence_header_offset
    ## End not in spec

    return video_parameters


@ref_pseudocode
def parse_parameters(state):
    """(11.2.1)"""
    state["major_version"] = read_uint(state)

    # (11.2.2) Check the major_version  is at least 1. (It may need to be
    # higher depending on the features used on the sequence, but this will be
    # checked as the stream is processed. Later, at the end of the stream we'll
    # also check the major_version was not too high for the set of features
    # actually used.)
    ## Begin not in spec
    if state["major_version"] < MINIMUM_MAJOR_VERSION:
        raise MajorVersionTooLow(state["major_version"])
    ## End not in spec

    state["minor_version"] = read_uint(state)

    # (11.2.2) Check the minor_version is 0.
    ## Begin not in spec
    if state["minor_version"] != 0:
        raise MinorVersionNotZero(state["minor_version"])
    ## End not in spec

    state["profile"] = read_uint(state)

    # (C.2) Profile must be a supported value
    assert_in_enum(state["profile"], Profiles, BadProfile)  ## Not in spec

    # (11.2.2) Profile must be supported by current version
    ## Begin not in spec
    minimum_required_version = profile_version_implication(state["profile"])
    if state["major_version"] < minimum_required_version:
        raise ProfileNotSupportedByVersion(state["profile"], state["major_version"])
    log_version_lower_bound(state, minimum_required_version)
    ## End not in spec

    state["level"] = read_uint(state)

    # (C.3) Level must be a supported value
    assert_in_enum(state["level"], Levels, BadLevel)  ## Not in spec

    # (C.3) Levels may constrain the order and choice of data units in a
    # sequence. See the various level standards documents (e.g. ST 2042-2) for
    # details.
    ## Begin not in spec
    if "_level_sequence_matcher" not in state:
        state["_level_sequence_matcher"] = Matcher(
            LEVEL_SEQUENCE_RESTRICTIONS[state["level"]].sequence_restriction_regex
        )
        # If we're at this point we're currently reading the first sequence
        # header (in the first data unit) of a sequence. Advance the state
        # machine accordingly.
        assert state["_level_sequence_matcher"].match_symbol("sequence_header")
    ## End not in spec

    # (C.3) Levels may constrain the allowed profiles and versions
    ## Begin not in spec
    assert_level_constraint(state, "level", state["level"])
    assert_level_constraint(state, "profile", state["profile"])
    assert_level_constraint(state, "major_version", state["major_version"])
    assert_level_constraint(state, "minor_version", state["minor_version"])
    ## End not in spec


@ref_pseudocode
def source_parameters(state, base_video_format):
    """(11.4.1)"""
    video_parameters = set_source_defaults(base_video_format)
    frame_size(state, video_parameters)
    color_diff_sampling_format(state, video_parameters)
    scan_format(state, video_parameters)
    frame_rate(state, video_parameters)
    pixel_aspect_ratio(state, video_parameters)
    clean_area(state, video_parameters)
    signal_range(state, video_parameters)
    color_spec(state, video_parameters)
    return video_parameters


@ref_pseudocode
def frame_size(state, video_parameters):
    """(11.4.3)"""
    custom_dimensions_flag = read_bool(state)
    # (C.3) Check level allows this value
    ## Begin not in spec
    assert_level_constraint(state, "custom_dimensions_flag", custom_dimensions_flag)
    ## End not in spec

    if custom_dimensions_flag:
        video_parameters["frame_width"] = read_uint(state)
        # (C.3) Check level allows this value
        ## Begin not in spec
        assert_level_constraint(state, "frame_width", video_parameters["frame_width"])
        ## End not in spec

        video_parameters["frame_height"] = read_uint(state)
        # (C.3) Check level allows this value
        ## Begin not in spec
        assert_level_constraint(state, "frame_height", video_parameters["frame_height"])
        ## End not in spec

        # Errata: spec doesn't prevent zero-pixel pictures
        #
        # (11.4.3) Frames must be at least one pixel wide and tall
        ## Begin not in spec
        if (
            video_parameters["frame_width"] == 0
            or video_parameters["frame_height"] == 0
        ):
            raise ZeroPixelFrameSize(
                video_parameters["frame_width"],
                video_parameters["frame_height"],
            )
        ## End not in spec


@ref_pseudocode
def color_diff_sampling_format(state, video_parameters):
    """(11.4.4)"""
    custom_color_diff_format_flag = read_bool(state)
    # (C.3) Check level allows custom color difference sampling
    ## Begin not in spec
    assert_level_constraint(
        state, "custom_color_diff_format_flag", custom_color_diff_format_flag
    )
    ## End not in spec

    if custom_color_diff_format_flag:
        video_parameters["color_diff_format_index"] = read_uint(state)

        # (11.4.4) Index shall be a known value
        ## Begin not in spec
        assert_in_enum(
            video_parameters["color_diff_format_index"],
            ColorDifferenceSamplingFormats,
            BadColorDifferenceSamplingFormat,
        )
        ## End not in spec

        # (C.3) Check level allows this value
        ## Begin not in spec
        assert_level_constraint(
            state,
            "color_diff_format_index",
            video_parameters["color_diff_format_index"],
        )
        ## End not in spec


@ref_pseudocode
def scan_format(state, video_parameters):
    """(11.4.5)"""
    custom_scan_format_flag = read_bool(state)
    # (C.3) Check level allows custom scan formats
    ## Begin not in spec
    assert_level_constraint(state, "custom_scan_format_flag", custom_scan_format_flag)
    ## End not in spec

    if custom_scan_format_flag:
        video_parameters["source_sampling"] = read_uint(state)

        # (11.4.5) Mode be a known value
        ## Begin not in spec
        assert_in_enum(
            video_parameters["source_sampling"],
            SourceSamplingModes,
            BadSourceSamplingMode,
        )
        ## End not in spec

        # (C.3) Check level allows this value
        ## Begin not in spec
        assert_level_constraint(
            state,
            "source_sampling",
            video_parameters["source_sampling"],
        )
        ## End not in spec


@ref_pseudocode
def frame_rate(state, video_parameters):
    """(11.4.6)"""
    custom_frame_rate_flag = read_bool(state)
    # (C.3) Check level allows custom frame rates
    ## Begin not in spec
    assert_level_constraint(state, "custom_frame_rate_flag", custom_frame_rate_flag)
    ## End not in spec

    if custom_frame_rate_flag:
        index = read_uint(state)

        # (C.3) Check level allows the specified preset
        assert_level_constraint(state, "frame_rate_index", index)  ## Not in spec

        if index == 0:
            video_parameters["frame_rate_numer"] = read_uint(state)
            # (C.3) Check level allows the specified numerator
            ## Begin not in spec
            assert_level_constraint(
                state,
                "frame_rate_numer",
                video_parameters["frame_rate_numer"],
            )
            ## End not in spec

            video_parameters["frame_rate_denom"] = read_uint(state)
            # (C.3) Check level allows the specified denominator
            ## Begin not in spec
            assert_level_constraint(
                state,
                "frame_rate_denom",
                video_parameters["frame_rate_denom"],
            )
            ## End not in spec

            # Errata: spec doesn't prevent divide-by-zero in custom frame rate
            # fractions.
            #
            # (11.4.6) frame_rate_denom must not be zero (i.e. a divide by zero)
            ## Begin not in spec
            if video_parameters["frame_rate_denom"] == 0:
                raise FrameRateHasZeroDenominator(video_parameters["frame_rate_numer"])
            ## End not in spec

            # Errata: spec doesn't prevent 0 fps
            #
            # (11.4.6) frame_rate_numer must not be zero (i.e. 0 fps)
            ## Begin not in spec
            if video_parameters["frame_rate_numer"] == 0:
                raise FrameRateHasZeroNumerator(video_parameters["frame_rate_denom"])
            ## End not in spec

        else:
            # (11.4.6) Frame rate preset must be a known value
            ## Begin not in spec
            assert_in_enum(index, PresetFrameRates, BadPresetFrameRateIndex)
            ## End not in spec

            # (11.2.2) Frame rate preset must be supported by current version
            ## Begin not in spec
            minimum_required_version = preset_frame_rate_version_implication(index)
            if state["major_version"] < minimum_required_version:
                raise PresetFrameRateNotSupportedByVersion(
                    index, state["major_version"]
                )
            log_version_lower_bound(state, minimum_required_version)
            ## End not in spec

            preset_frame_rate(video_parameters, index)


@ref_pseudocode
def pixel_aspect_ratio(state, video_parameters):
    """(11.4.7)"""
    custom_pixel_aspect_ratio_flag = read_bool(state)
    # (C.3) Check level allows custom pixel aspect ratio
    ## Begin not in spec
    assert_level_constraint(
        state, "custom_pixel_aspect_ratio_flag", custom_pixel_aspect_ratio_flag
    )
    ## End not in spec

    if custom_pixel_aspect_ratio_flag:
        index = read_uint(state)
        # (C.3) Check level allows the specified preset
        ## Begin not in spec
        assert_level_constraint(state, "pixel_aspect_ratio_index", index)
        ## End not in spec

        if index == 0:
            video_parameters["pixel_aspect_ratio_numer"] = read_uint(state)
            # (C.3) Check level allows the specified numerator
            ## Begin not in spec
            assert_level_constraint(
                state,
                "pixel_aspect_ratio_numer",
                video_parameters["pixel_aspect_ratio_numer"],
            )
            ## End not in spec

            video_parameters["pixel_aspect_ratio_denom"] = read_uint(state)
            # (C.3) Check level allows the specified denominator
            ## Begin not in spec
            assert_level_constraint(
                state,
                "pixel_aspect_ratio_denom",
                video_parameters["pixel_aspect_ratio_denom"],
            )
            ## End not in spec

            # Errata: spec fails to require ratio to be non-zero on either side
            #
            # (11.4.7) ratio must not contain zeros
            ## Begin not in spec
            if (
                video_parameters["pixel_aspect_ratio_numer"] == 0
                or video_parameters["pixel_aspect_ratio_denom"] == 0
            ):
                raise PixelAspectRatioContainsZeros(
                    video_parameters["pixel_aspect_ratio_numer"],
                    video_parameters["pixel_aspect_ratio_denom"],
                )
            ## End not in spec
        else:
            # (11.4.7) Pixel aspect ratio preset must be a known value
            ## Begin not in spec
            assert_in_enum(index, PresetPixelAspectRatios, BadPresetPixelAspectRatio)
            ## End not in spec

            preset_pixel_aspect_ratio(video_parameters, index)


@ref_pseudocode
def clean_area(state, video_parameters):
    """(11.4.8)"""
    custom_clean_area_flag = read_bool(state)
    # (C.3) Check level allows custom clean area
    ## Begin not in spec
    assert_level_constraint(state, "custom_clean_area_flag", custom_clean_area_flag)
    ## End not in spec

    if custom_clean_area_flag:
        video_parameters["clean_width"] = read_uint(state)
        # (C.3) Check level allows this width
        ## Begin not in spec
        assert_level_constraint(state, "clean_width", video_parameters["clean_width"])
        ## End not in spec

        video_parameters["clean_height"] = read_uint(state)
        # (C.3) Check level allows this height
        ## Begin not in spec
        assert_level_constraint(state, "clean_height", video_parameters["clean_height"])
        ## End not in spec

        video_parameters["left_offset"] = read_uint(state)
        # (C.3) Check level allows this offset
        ## Begin not in spec
        assert_level_constraint(state, "left_offset", video_parameters["left_offset"])
        ## End not in spec

        video_parameters["top_offset"] = read_uint(state)
        # (C.3) Check level allows this offset
        ## Begin not in spec
        assert_level_constraint(state, "top_offset", video_parameters["top_offset"])
        ## End not in spec

    # (11.4.8) The clean area is restricted to being within the existing
    # picture area
    ## Begin not in spec
    if not (
        video_parameters["clean_width"] + video_parameters["left_offset"]
        <= video_parameters["frame_width"]
        and video_parameters["clean_height"] + video_parameters["top_offset"]
        <= video_parameters["frame_height"]
    ):
        raise CleanAreaOutOfRange(
            video_parameters["clean_width"],
            video_parameters["clean_height"],
            video_parameters["left_offset"],
            video_parameters["top_offset"],
            video_parameters["frame_width"],
            video_parameters["frame_height"],
        )
    ## End not in spec


@ref_pseudocode
def signal_range(state, video_parameters):
    """(11.4.9)"""
    custom_signal_range_flag = read_bool(state)
    # (C.3) Check level allows custom signal range
    ## Begin not in spec
    assert_level_constraint(state, "custom_signal_range_flag", custom_signal_range_flag)
    ## End not in spec

    if custom_signal_range_flag:
        index = read_uint(state)
        # (C.3) Check level allows the specified preset
        ## Begin not in spec
        assert_level_constraint(state, "custom_signal_range_index", index)
        ## End not in spec

        if index == 0:
            video_parameters["luma_offset"] = read_uint(state)
            # (C.3) Check level allows this offset
            ## Begin not in spec
            assert_level_constraint(
                state, "luma_offset", video_parameters["luma_offset"]
            )
            ## End not in spec

            video_parameters["luma_excursion"] = read_uint(state)
            # (C.3) Check level allows this excursion
            ## Begin not in spec
            assert_level_constraint(
                state, "luma_excursion", video_parameters["luma_excursion"]
            )
            ## End not in spec

            # (11.4.9) Check luma_excursion is valid
            #
            # Errata: Spec fails to constrain excursions.  Excursions *must* be
            # >= 1 or the pseudocode behaviour will be undefined.
            ## Begin not in spec
            if video_parameters["luma_excursion"] < 1:
                raise BadCustomSignalExcursion(
                    "luma", video_parameters["luma_excursion"]
                )
            ## End not in spec

            video_parameters["color_diff_offset"] = read_uint(state)
            # (C.3) Check level allows this offset
            ## Begin not in spec
            assert_level_constraint(
                state, "color_diff_offset", video_parameters["color_diff_offset"]
            )
            ## End not in spec

            video_parameters["color_diff_excursion"] = read_uint(state)
            # (C.3) Check level allows this excursion
            ## Begin not in spec
            assert_level_constraint(
                state, "color_diff_excursion", video_parameters["color_diff_excursion"]
            )
            ## End not in spec

            # (11.4.9) Check color_diff_excursion is valid
            #
            # Errata: Spec fails to constrain excursions.  Excursions *must* be
            # >= 1 or the pseudocode behaviour will be undefined.
            ## Begin not in spec
            if video_parameters["color_diff_excursion"] < 1:
                raise BadCustomSignalExcursion(
                    "color_diff", video_parameters["color_diff_excursion"]
                )
            ## End not in spec
        else:
            # (11.4.9) Signal range preset must be a known value
            ## Begin not in spec
            assert_in_enum(index, PresetSignalRanges, BadPresetSignalRange)
            ## End not in spec

            # (11.2.2) Signal range preset must be supported by current version
            ## Begin not in spec
            minimum_required_version = preset_signal_range_version_implication(index)
            if state["major_version"] < minimum_required_version:
                raise PresetSignalRangeNotSupportedByVersion(
                    index, state["major_version"]
                )
            log_version_lower_bound(state, minimum_required_version)
            ## End not in spec

            preset_signal_range(video_parameters, index)


@ref_pseudocode
def color_spec(state, video_parameters):
    """(11.4.10.1)"""
    custom_color_spec_flag = read_bool(state)
    # (C.3) Check level allows custom color specs
    ## Begin not in spec
    assert_level_constraint(state, "custom_color_spec_flag", custom_color_spec_flag)
    ## End not in spec

    if custom_color_spec_flag:
        index = read_uint(state)

        # (11.4.10.1) Index should be a supported value
        assert_in_enum(index, PresetColorSpecs, BadPresetColorSpec)  ## Not in spec

        # (C.3) Check level allows the specified preset
        assert_level_constraint(state, "color_spec_index", index)  ## Not in spec

        preset_color_spec(video_parameters, index)

        if index == 0:
            color_primaries(state, video_parameters)
            color_matrix(state, video_parameters)
            transfer_function(state, video_parameters)
        ## Begin not in spec
        else:
            # (11.2.2) Color spec must be supported by current version
            minimum_required_version = preset_color_spec_version_implication(index)
            if state["major_version"] < minimum_required_version:
                raise PresetColorSpecNotSupportedByVersion(
                    index, state["major_version"]
                )
            log_version_lower_bound(state, minimum_required_version)
        ## End not in spec


@ref_pseudocode
def color_primaries(state, video_parameters):
    """(11.4.10.2)"""
    custom_color_primaries_flag = read_bool(state)
    # (C.3) Check level allows custom color primaries
    ## Begin not in spec
    assert_level_constraint(
        state, "custom_color_primaries_flag", custom_color_primaries_flag
    )
    ## End not in spec

    if custom_color_primaries_flag:
        index = read_uint(state)

        # (11.4.10.2) Index should be a supported value
        ## Begin not in spec
        assert_in_enum(index, PresetColorPrimaries, BadPresetColorPrimaries)
        ## End not in spec

        # (C.3) Check level allows the specified preset
        ## Begin not in spec
        assert_level_constraint(state, "color_primaries_index", index)
        ## End not in spec

        # (11.2.2) Preset must be supported by current version
        ## Begin not in spec
        minimum_required_version = preset_color_primaries_version_implication(index)
        if state["major_version"] < minimum_required_version:
            raise PresetColorPrimariesNotSupportedByVersion(
                index, state["major_version"]
            )
        log_version_lower_bound(state, minimum_required_version)
        ## End not in spec

        preset_color_primaries(video_parameters, index)


@ref_pseudocode
def color_matrix(state, video_parameters):
    """(11.4.10.3)"""
    custom_color_matrix_flag = read_bool(state)
    # (C.3) Check level allows custom color matrices
    ## Begin not in spec
    assert_level_constraint(state, "custom_color_matrix_flag", custom_color_matrix_flag)
    ## End not in spec

    if custom_color_matrix_flag:
        index = read_uint(state)

        # (11.4.10.3) Index should be a supported value
        assert_in_enum(index, PresetColorMatrices, BadPresetColorMatrix)  ## Not in spec

        # (C.3) Check level allows the specified preset
        ## Begin not in spec
        assert_level_constraint(state, "color_matrix_index", index)
        ## End not in spec

        # (11.2.2) Preset must be supported by current version
        ## Begin not in spec
        minimum_required_version = preset_color_matrix_version_implication(index)
        if state["major_version"] < minimum_required_version:
            raise PresetColorMatrixNotSupportedByVersion(index, state["major_version"])
        log_version_lower_bound(state, minimum_required_version)
        ## End not in spec

        preset_color_matrix(video_parameters, index)


@ref_pseudocode
def transfer_function(state, video_parameters):
    """(11.4.10.4)"""
    custom_transfer_function_flag = read_bool(state)
    # (C.3) Check level allows custom transfer functions
    ## Begin not in spec
    assert_level_constraint(
        state, "custom_transfer_function_flag", custom_transfer_function_flag
    )
    ## End not in spec

    if custom_transfer_function_flag:
        index = read_uint(state)

        # (11.4.10.3) Index should be a supported value
        ## Begin not in spec
        assert_in_enum(index, PresetTransferFunctions, BadPresetTransferFunction)
        ## End not in spec

        # (C.3) Check level allows the specified preset
        ## Begin not in spec
        assert_level_constraint(state, "transfer_function_index", index)
        ## End not in spec

        # (11.2.2) Preset must be supported by current version
        ## Begin not in spec
        minimum_required_version = preset_transfer_function_version_implication(index)
        if state["major_version"] < minimum_required_version:
            raise PresetTransferFunctionNotSupportedByVersion(
                index, state["major_version"]
            )
        log_version_lower_bound(state, minimum_required_version)
        ## End not in spec

        preset_transfer_function(video_parameters, index)



r"""
.. _vc2-bitstream-viewer:

``vc2-bitstream-viewer``
========================

A command-line utility for displaying VC-2 bitstreams in a human readable form.

The following tutorials give a higher-level introduction to this tool.
Complete usage information can be obtained by running ``vc2-bitstream-viewer
--help``.

.. warning::

    This program prioritises correctness over performance. Internally the
    pseudocode from the VC-2 specification is used to drive the bitstream
    parser. As a consequence this program is relatively slow, e.g. taking many
    seconds per HD frame.


Basic usage
-----------

In its simplest form, this command can be used to create a complete textual
representation of a VC-2 bitstream::

    $ vc2-bitstream-viewer path/to/bitstream.vc2
                                                      +- data_units:
                                                      | +- 0:
                                                      | | +- parse_info:
    000000000000:                                     | | | +- padding: 0b
    000000000000: 01000010010000100100001101000100    | | | +- parse_info_prefix: Correct (0x42424344)
    000000000032: 00010000                            | | | +- parse_code: end_of_sequence (0x10)
    000000000040: 00000000000000000000000000000000    | | | +- next_parse_offset: 0
    000000000072: 00000000000000000000000000000000    | | | +- previous_parse_offset: 0

In the printed output, each value read from the bitstream is shown on its own
line:

* The number at the start of the line is the offset (in bits) from the
  start of the file that that value begins.
* The next value is the value in the bitstream expressed as a binary number.
  If values are being read from the bounded block in the bitstream, a (``*``)
  indicates that any remaining bits in the value shown were read from beyond
  the end of the block (and read as '1').
* The final part of the line indicates the name and decoded value of the read
  data. Where possible, the names used match the names of variables in the VC-2
  pseudocode.

By default, every bit and every field in the bitstream will be visible,
including padding bits.


Displaying an area of interest
------------------------------

It is possible to restrict the output to only parts of the bitstream
surrounding a particular offset. For example, the ``--offset``/``-o`` argument
can be used to restrict display to only the parts of the bitstream surrounding
a particular bit offset::

    $ vc2-bitstream-viewer bitstream.vc2 --offset 40000

See the ``--context``/``-C``, ``--context-after``/``-A`` and
``--context-before``/``-B`` arguments to control the number of bits before and
after the supplied offset to be shown. Alternatively, you can also specify an
explicit range to display using the ``--from-offset``/``-f`` and
``--to-offset``/``-t``.

.. note::

    Though values before the specified bit offset will not be displayed, they
    must still be read and parsed by the bitstream viewer in order to correctly
    parse what comes later in the bitstream.


Filtering displayed values
--------------------------

It is possible to filter the displayed values according to the VC-2 pseudocode
function which read them. A common case might be to show only the parse_info
headers from a stream::

    $ vc2-bitstream-viewer bitstream.vc2 --show parse_info

Alternatively, you might wish to show everything in a bitstream except for the
transform coefficients (slice data)::

    $ vc2-bitstream-viewer bitstream.vc2 --hide slice

For this particular action, the convenience alias ``-S`` for ``--hide slice``
is also provided.

Showing VC-2 decoder state
--------------------------

As well as displaying a textual representation of the VC-2 bitstream, this tool
can also display (a subset of) the internal state of a VC-2 decoder based on
the pseudocode in the specification at different points during bitstream
processing. Use the ``--show-internal-state`` option to enable this. This can
be useful when debugging the encoding of transform data whose organisation
depends on various computed values earlier in the bitstream.


Malformed bitstream handling
----------------------------

The bitstream parser is as tolerant of malformed bitstreams as is practical and
will accept, and blindly tolerate values which are out-of-spec so long as they
do not lead to undefined behaviour. In the event that the parser encounters an
anomaly that prevents it preceding further, an error will be shown and parsing
will terminate.

By default, error messages show only limited information about the failure.
Adding the ``--verbose``/``-v`` option will cause additional details (such as
the next few bits in the bitstream and details of what was being parsed at the
time.

Arguments
---------

The complete set of arguments can be listed using ``--help``

.. program-output:: vc2-bitstream-viewer --help

"""  # noqa: E501

import os
import sys
import time
import inspect
import traceback

from bitarray import bitarray

from argparse import ArgumentParser

from textwrap import wrap

from vc2_conformance import __version__

from vc2_conformance.pseudocode.metadata import make_pseudocode_traceback

from vc2_conformance import bitstream

from vc2_conformance.pseudocode.state import State

from vc2_data_tables import PARSE_INFO_PREFIX

from vc2_conformance.string_utils import ellipsise_lossy
from vc2_conformance.string_formatters import Hex

from vc2_conformance.py2x_compat import (
    zip_longest,
    get_terminal_size,
)


DEFAULT_CONTEXT_BITS = 128
"""Number of bits either side of --offset to show by default."""

STATUS_LINE_UPDATE_INTERVAL = 0.2
"""Number of seconds between updates to the status line."""

OFFSET_DIGITS = 12
"""Number of digits to show in bit offsets"""

RAW_BITS_PER_LINE = 32
"""Number of binary digits to show per line in raw binary strings."""


def relative_int(string):
    """
    Given an integer of the form accepted by --to-offset, return a (relative,
    offset) pair. 'relative' will be True for relative offsets and False
    otherwise. 'offset' will be an int.
    """
    if not isinstance(string, str):
        raise ValueError(string)

    relative = string.startswith("+")
    offset = int(string)

    return (relative, offset)


def relative_to_abs_index(num, length):
    """
    Given a relative index (negative values are relative to the end of the
    length), return an absolute index.
    """
    if num >= 0:
        return num
    else:
        return length + num


# The filename of this module and the bitstream VC2 module (not the .pyc files)
_this_script_filename = inspect.getsourcefile(sys.modules[__name__])
_bitstream_vc2_filename = inspect.getsourcefile(bitstream.vc2)


def is_internal_error(tb):
    """
    Given a Python traceback, attempt to determine if the fault lies with a bug
    in this script (i.e. not an error produced by the VC-2 pseudo code
    encountering an unnacceptably out-of-range value).
    """
    is_internal = False

    stack_summary = traceback.extract_tb(tb)
    for frame_summary in stack_summary:
        filename = frame_summary[0]

        if filename == _this_script_filename:
            is_internal = True
        elif filename == _bitstream_vc2_filename:
            is_internal = False

    return is_internal


def _call(f, *args, **kwargs):
    """
    For use by the test suite. Calls function 'f' with the provided arguments.
    """
    return f(*args, **kwargs)


def most_recent_pseudocode_function(tb):
    """
    Given a Python traceback, return the function name/citation of the
    inner-most VC-2 pseudocode function to be executed.
    """
    ptb = make_pseudocode_traceback(traceback.extract_tb(tb))
    if ptb:
        return ptb[-1].citation
    else:
        return "<unknown>"


def format_path_summary(path):
    """
    Format a :py:mod:`serdes` path as a string.
    """
    return ": ".join(map(str, path))


def format_value_line(offset, raw_bits, label, label_filler="", truncated=False):
    """
    Format a line of output which displays a single value from the bitstream.

    Parameters
    ----------
    offset : int
        Bit offset into bitstream.
    raw_bits : :py:class:`bitarray.bitarray`
        The raw bits which encode this value.
    label : str
        The label which describes the value encoded by the raw_bits.
    label_filler : str
        If the raw_bits, when printed, spread onto multiple lines, the string
        to show in place of the label on these lines.
    truncated : bool
        If True, shows a '*' at the end of the raw_bits to indicate that
        truncation (e.g. at the end of a bounded block) has taken place.
    """
    raw_bits_suffix = "*" if truncated else ""
    raw_bits_lines = wrap(raw_bits.to01() + raw_bits_suffix, RAW_BITS_PER_LINE)

    if len(raw_bits_lines) == 0:
        raw_bits_lines = [""]

    out = []
    out.append(
        "{:0{}d}: {:<{}s}    {}".format(
            offset,
            OFFSET_DIGITS,
            raw_bits_lines[0],
            RAW_BITS_PER_LINE,
            label,
        )
    )

    for line in raw_bits_lines[1:]:
        out.append(
            "{}  {:<{}s}    {}".format(
                " " * OFFSET_DIGITS, line, RAW_BITS_PER_LINE, label_filler
            )
        )

    return "\n".join(out)


def format_header_line(label):
    """
    Format a line of output which displays only a header (indented to match the
    labels printed by :py:func:`format_value_line`.
    """
    return "{}{}".format(
        " " * (OFFSET_DIGITS + 2 + RAW_BITS_PER_LINE + 4),
        label,
    )


def format_omission_line(offset, num_bits):
    """
    Format a line of output which indicates a number of bits have been omitted
    from the bitstream.
    """
    return "{:0{}d}: {:<{}s}    ...".format(
        offset,
        OFFSET_DIGITS,
        "<{} bits omitted>".format(num_bits),
        RAW_BITS_PER_LINE,
    )


class BitstreamViewer(object):
    """
    Main commandline application logic.

    Construct an instance of this class then call :py:meth:`run` and, later,
    :py:meth:`close`.
    """

    def __init__(
        self,
        filename,
        from_offset=0,
        to_offset=-1,
        shown_pseudocode_names=[],
        hidden_pseudocode_names=[],
        show_internal_state=False,
        check_parse_info_prefix=True,
        show_status=True,
        verbose=0,
        num_trailing_bits_on_error=128,
    ):
        """
        Parameters
        ==========
        filename : str
            The filename of the bitstream.
        from_offset : int
            The bit offset at which to start reading from the bitstream. If
            negative, will be assumed to be an offset from the end of the file.
        to_offset : int
            The bit offset at which to stop reading the bitstream. If negative,
            will be assumed to be an offset from the end of the file.
        shown_pseudocode_names : [str, ...]
            A list of strings naming the bitstream pseudocode functions whose
            values should be shown in the printed output. If empty, all values
            are shown.
        hidden_pseudocode_names : [str, ...]
            A list of strings naming the bitstream pseudocode functions whose
            values should be omitted in the printed output.
        show_internal_state : bool
            If True, the internal state of the pseudocode functions will be
            printed before and after every data unit.
        check_parse_info_prefix : bool
            If True, validate parse_info prefix bytes and stop if they don't
            hold the expected value.
        show_status : bool
            If True, show a status line on stderr while reading parts of the
            bitstream which aren't being printed.
        verbose : int
            Error message verbosity level:

            0. Only show the error message.
            1. Show trailing bits in bitstream, bit offset and current
               target.
            2. Show traceback.
        num_trailing_bits_on_error : int
            When ``verbose`` is at least 1, controls the (maximum) number of
            bits in the bitstream to display after an error has ocurred.
        """
        self._filename = filename
        self._from_offset = from_offset
        self._to_offset = to_offset
        self._show_internal_state = show_internal_state
        self._check_parse_info_prefix = check_parse_info_prefix
        self._show_status = show_status
        self._verbose = verbose
        self._num_trailing_bits_on_error = num_trailing_bits_on_error

        # A set of fixeddict types which are to be shown or hidden (None if no
        # filter).
        self._shown_types = None
        self._hidden_types = None

        # Convert from pseudocode function names to sets of corresponding
        # fixeddict types for the above
        shown = set()
        for name in shown_pseudocode_names:
            shown.update(bitstream.pseudocode_function_to_fixeddicts_recursive[name])
        hidden = set()
        for name in hidden_pseudocode_names:
            hidden.update(bitstream.pseudocode_function_to_fixeddicts_recursive[name])
        if len(shown) > 0:
            self._shown_types = shown - hidden
        elif len(hidden) > 0:
            self._hidden_types = hidden

        # The file object for the bitstream
        self._file = None

        # The BitstreamReader reading the file
        self._reader = None

        # The MonitoredDeserialiser in use
        self._serdes = None

        # The internal State variable used by VC-2 pseudocode
        self._state = State()

        # Is the status line currently visible
        self._status_line_visible = False

        # The last time at which the status bar was updated
        self._status_line_updated = 0.0

        # The bitstream offset last time the monitor function was called
        self._last_tell = (0, 7)

        # The last bitstream offset at which the monitor function printed a
        # value
        self._last_displayed_tell = (0, 7)

        # The last serdes path displayed
        self._last_path = []

        # The last number of data units in the stream (only used when
        # show_internal_state is True)
        self._last_num_data_units = 0

    def _print_error(self, message):
        """
        Print an error message to stderr.
        """
        # Avoid interleaving with stdout (and make causality clearer)
        sys.stdout.flush()

        # The 'trailing bits' print out below might move the read position
        # so record it now.
        if self._reader is not None:
            offset = bitstream.to_bit_offset(*self._reader.tell())

        # Display the trailing bits from the bitstream
        if self._verbose >= 1:
            if self._reader is not None:
                self._reader.seek(*self._last_tell)
                bits = self._reader.try_read_bitarray(self._num_trailing_bits_on_error)
                if len(bits) > 0:
                    sys.stderr.write(
                        "{}\n".format(
                            format_value_line(
                                bitstream.to_bit_offset(*self._last_tell),
                                bitarray(bits),
                                "<next {} bits>".format(len(bits)),
                            )
                        )
                    )

        # Display the traceback
        if self._verbose >= 2:
            if sys.exc_info()[0] is not None:
                traceback.print_exc()

        # Display offset/target information
        prog = os.path.basename(sys.argv[0])
        if self._verbose >= 1:
            if self._reader is not None:
                sys.stderr.write(
                    "{}: offset: {}\n".format(
                        prog,
                        offset,
                    )
                )
            if self._serdes is not None:
                sys.stderr.write(
                    "{}: target: {}\n".format(
                        prog,
                        format_path_summary(self._serdes.path()),
                    )
                )

        # Display the message
        message = "{}: error: {}".format(prog, message)
        sys.stderr.write("{}\n".format(message))

    class _TerminateSuccess(Exception):
        """
        Thrown from the monitoring function to indicate a healty early
        termination of bitstream parsing.
        """

        pass

    class _TerminateError(Exception):
        """
        Thrown from the monitoring function to indicate a fatal error during
        bitstream parsing..
        """

        pass

    def _update_status_line(self, target):
        """
        Update the status line, if a sufficiently long time has ellpased (to
        avoid spending too long printing to the console).
        """
        now = time.time()
        if now - self._status_line_updated >= STATUS_LINE_UPDATE_INTERVAL:
            self._status_line_visible = True
            self._status_line_updated = now

            cur_offset = bitstream.to_bit_offset(*self._reader.tell())
            cur_path = format_path_summary(self._serdes.path(target))

            terminal_width = get_terminal_size()[0]
            line = "{:0{}d}: <{}>".format(
                cur_offset,
                OFFSET_DIGITS,
                ellipsise_lossy(cur_path, terminal_width - OFFSET_DIGITS - 4),
            )

            # Ensure stdout is fully displayed before printing the status line
            sys.stdout.flush()

            sys.stderr.write(
                (
                    "\033[2K"  # Clear to end of line
                    "\033[s"  # Save cursor position
                    "{}"
                    "\033[u"  # Restore cursor position
                ).format(line)
            )
            sys.stderr.flush()

    def _hide_status_line(self):
        """If the status line is visible, hide it."""
        if self._status_line_visible:
            self._status_line_visible = False

            # Ensure stdout is fully displayed before doing anything to the
            # status line.
            sys.stdout.flush()

            sys.stderr.write("\033[2K")  # Clear to end of line
            sys.stderr.flush()

    def _print_value(self, offset, raw_bits, target, value):
        """
        Print a value from the bitstream on stdout.

        Parameters
        ==========
        offset : int
            The bit offset at the start of the value in the bitstraem.
        raw_bits : bitarray
            The raw bits read from the bitstream for this value.
        target : str
            The target name of this value.
        value : any
            The deserialised value.
        """
        # Print new/changed path prefixes
        last_path = self._last_path
        this_path = self._serdes.path(target)
        self._last_path = this_path
        different = False
        for level, (last_path_element, this_path_element) in enumerate(
            zip_longest(last_path[:-1], this_path[:-1])
        ):
            different |= (
                last_path_element != this_path_element and this_path_element is not None
            )
            if different and this_path_element is not None:
                prefix = "| " * level
                print(format_header_line("{}+- {}:".format(prefix, this_path_element)))

        # Mark values read past the end of a bounded block
        past_end_of_bounded_block = (
            self._reader.bits_remaining is not None and self._reader.bits_remaining < 0
        )

        # Format value for display
        if hasattr(self._serdes.cur_context, "entry_objs"):
            formatter = self._serdes.cur_context.entry_objs[target].to_string
        else:
            formatter = str
        if self._serdes.cur_context[target] is not value:
            # This is a list target, use the formatter's inner formatter
            formatter = getattr(formatter, "formatter", str)

        value_prefix = "{}+- {}: ".format("| " * (len(this_path) - 1), this_path[-1])
        value_str = "{}{}".format(value_prefix, formatter(value))

        value_filler_str = "| " * (len(this_path))

        print(
            format_value_line(
                offset,
                raw_bits,
                value_str,
                value_filler_str,
                past_end_of_bounded_block,
            )
        )

    def _print_internal_state(self):
        """
        Display the current internal state of the VC-2 pseudocode functions.
        """
        if self._state:
            print("-" * (OFFSET_DIGITS + 2 + RAW_BITS_PER_LINE))
            print(str(self._state))
            print("-" * (OFFSET_DIGITS + 2 + RAW_BITS_PER_LINE))

    def _print_omitted_bits(self, tell):
        """
        Display the number of bits omitted since the last displayed value up
        until 'tell'.
        """
        offset = bitstream.to_bit_offset(*tell)
        last_displayed_offset = bitstream.to_bit_offset(*self._last_displayed_tell)
        num_bits = offset - last_displayed_offset

        print(format_omission_line(last_displayed_offset, num_bits))

    def __call__(self, serdes, target, value):
        """
        Called as the monitoring function of the the
        :py:class:`MonitoredDeserialiser`. Responsible for printing values from
        the bitstream and terminating once the region of interest has been
        parsed.
        """
        last_tell = self._last_tell
        this_tell = self._reader.tell()
        self._last_tell = this_tell

        if self._show_internal_state:
            # Only show at start of new data unit (NB: strictly speaking this
            # will display the state was it was after parsing the first field
            # of each data unit. Since this is a padding field, the state
            # should still be correct.
            num_data_units = sum(
                len(sequence["data_units"])
                for sequence in self._serdes.context["sequences"]
            )
            if self._last_num_data_units != num_data_units:
                self._last_num_data_units = num_data_units

                self._hide_status_line()
                self._print_internal_state()

        this_offset = bitstream.to_bit_offset(*this_tell)

        enable_display = False
        enable_display = (
            # Within the specified region (NB: we terminate once past it)
            (self._from_offset == 0 or this_offset > self._from_offset)
            and
            # Is the current context selected to be shown
            (
                self._hidden_types is None
                or self._serdes.cur_context.__class__ not in self._hidden_types
            )
            and (
                self._shown_types is None
                or self._serdes.cur_context.__class__ in self._shown_types
            )
        )

        if enable_display:
            self._hide_status_line()

            last_offset = bitstream.to_bit_offset(*last_tell)

            # Print message if we've skipped some bits
            if self._last_displayed_tell != last_tell:
                self._print_omitted_bits(last_tell)
            self._last_displayed_tell = this_tell

            # Re-read the bits within the value just read and display it
            self._reader.seek(*last_tell)
            raw_bits = self._reader.read_bitarray(this_offset - last_offset)
            self._print_value(last_offset, raw_bits, target, value)
        else:
            if self._show_status:
                self._update_status_line(target)

        # Stop if we've encountered an invalid parse_info prefix code
        if self._check_parse_info_prefix:
            # The test below relies on the only use of the target name
            # 'parse_info_prefix' being for the parse info prefix. (This should
            # be the case).
            if target == "parse_info_prefix" and value != PARSE_INFO_PREFIX:
                raise BitstreamViewer._TerminateError(
                    "invalid parse_info prefix ({})".format(Hex(8)(value))
                )

        # Stop when we've reached/passed the end of the region of interest
        if self._to_offset != 0 and this_offset >= self._to_offset:
            raise BitstreamViewer._TerminateSuccess()

        # Save memory by discarding previously deserialised data units
        current_data_unit = True
        for sequence in reversed(self._serdes.context["sequences"]):
            for i in reversed(range(len(sequence["data_units"]))):
                if not current_data_unit:
                    sequence["data_units"][i] = None
                current_data_unit = False

    def run(self):
        """
        Parse the bitstream. Returns 0 if the bitstream was read successfully
        or another integer otherwise.
        """
        # Open the file
        try:
            self._file = open(self._filename, "rb")
            self._reader = bitstream.BitstreamReader(self._file)
            filesize_bytes = os.path.getsize(self._filename)
        except Exception as e:
            # Catch-all exception handler excuse: Catching only file-related
            # exceptions is challenging, particularly in a backward-compatible
            # manner. However, none of the above are known to produce
            # exceptions *except* due to file-related issues.
            self._print_error(str(e))
            return 1

        # Resolve filesizes to absolute values
        filesize = filesize_bytes * 8
        self._from_offset = relative_to_abs_index(self._from_offset, filesize)
        self._to_offset = max(0, relative_to_abs_index(self._to_offset, filesize))

        return_code = 0
        error_message = None

        # Parse the bitstream. A MonitoredDeserialiser is used which calls
        # __call__ whenever a value is read. That method is responsible for
        # printing the bitstream to stdout and also deciding when to terminate
        # the read process, raising _TerminateSuccess and _TerminateError to
        # end parsing before the end of the file.
        try:
            self._serdes = bitstream.MonitoredDeserialiser(
                io=self._reader,
                monitor=self,
            )
            bitstream.parse_stream(self._serdes, self._state)
        except BitstreamViewer._TerminateSuccess:
            return_code = 0
            error_message = None
        except KeyboardInterrupt:
            return_code = 1
            error_message = None
        except BitstreamViewer._TerminateError as e:
            return_code = 2
            error_message = str(e)
        except EOFError:
            return_code = 3
            error_message = "reached the end of the file while parsing {}".format(
                most_recent_pseudocode_function(sys.exc_info()[2]),
            )
        except Exception:
            # Other exceptions might be raised during parsing (e.g. due to
            # invalid bitstream values), attempt to display these sensibly.
            exc_type, exc_value, exc_tb = sys.exc_info()
            if is_internal_error(exc_tb):
                # If the exception does not originate from the VC-2 pseudocode,
                # we've encountered an internal error in this program. This
                # should not happen, but if it does it should be made clear.
                return_code = 255
                error_message = (
                    "internal error in bitstream viewer: {}: {} "
                    "(probably a bug in this program)".format(
                        exc_type.__name__,
                        str(exc_value),
                    )
                )
            else:
                # General case: some error in the VC-2 pseudocode due to an
                # out-of-range value
                return_code = 4
                error_message = (
                    "{} failed to parse bitstream ({}: {}) "
                    "(missing sequence_header, fragment or earlier out of range value?)"
                ).format(
                    most_recent_pseudocode_function(exc_tb),
                    exc_type.__name__,
                    str(exc_value),
                )
        finally:
            self._hide_status_line()

        if self._last_displayed_tell != self._last_tell:
            self._print_omitted_bits(self._last_tell)

        if self._show_internal_state:
            self._print_internal_state()

        if error_message is not None:
            self._print_error(error_message)

        return return_code

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None


def parse_args(*args, **kwargs):
    """
    Parse a set of command line arguments. Returns a :py:mod:`argparse`
    ``args`` object with the following fields:

    * bitstream (str): The filename of the bitstream to read
    * no_status (bool): True if the status line is to be hidden.
    * verbose (int): Verbosity level
    * show_internal_state (bool): True if the state of the VC-2 pseudo code
      should be shown during parsing.
    * ignore_parse_info_prefix (bool): True if the parse_info prefix code
      should not be checked for correctness.
    * num_trailing_bits (int): Number of unprocessed bits from the bitstream to
      display in verbose error messages.
    * from_offset (int): The number of bits after the start of file to reach
      before displaying output. If negative, relative to the EOF.
    * to_offset (int): The bit offset at which to stop printing. If negative,
      relative to the EOF.
    * show (list of str): List of VC-2 pseudocode function names whose
      bitstream values must be displayed in the output. If empty, show
      everything.
    * hide (list of str): List of VC-2 pseudocode function names whose
      bitstream values must be hidden in the output.
    """
    parser = ArgumentParser(
        description="""
        Display VC-2 bitstreams in a human-readable form.
    """
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )

    parser.add_argument(
        "bitstream",
        help="""
            The filename of the bitstream to read.
        """,
    )

    parser.add_argument(
        "--no-status",
        "--quiet",
        "-q",
        action="store_true",
        default=False,
        help="""
            Do not display a status line on stderr while reading the bitstream.
        """,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="""
            Increase the verbosity of error messages. Used once: also show bitstream
            offset, bitstream target value and next --num-trailing-bits bits in
            the bitstream at the time of the error. Used twice: also show the
            Python stack trace for the error.
        """,
    )

    parser.add_argument(
        "--show-internal-state",
        "-i",
        action="store_true",
        default=False,
        help="""
            Print the internal state variable used by the VC-2 pseudo code
            after parsing each data unit. Parts of the pseudocode state not
            directly related to bitstream processing might or might not be
            included in this printout.
        """,
    )

    parser.add_argument(
        "--ignore-parse-info-prefix",
        "-p",
        action="store_true",
        default=False,
        help="""
            By default, parsing is halted if an invalid parse_info prefix is
            encountered. Giving this option suppresses this check and allows
            parsing to continue.
        """,
    )

    parser.add_argument(
        "--num-trailing-bits",
        "-b",
        type=int,
        default=DEFAULT_CONTEXT_BITS,
        help="""
            When --verbose is used, this argument defines the number of
            bits to show from the bitstream after the point the error occurred.
            (Default: %(default)d).
        """,
    )

    ###########################################################################

    range_group = parser.add_argument_group(title="range options")

    range_group.add_argument(
        "--from-offset",
        "-f",
        type=int,
        metavar="BIT_OFFSET",
        help="""
            Don't display bitstream values until the specified bit offset. If
            negative, gives an offset from the end of the file. Default: 0.
        """,
    )

    range_group.add_argument(
        "--to-offset",
        "-t",
        metavar="BIT_OFFSET",
        type=relative_int,
        help="""
            Stop reading the bitstream at the specified bit offset. If prefixed
            with '+', the offset will be relative to the offset given to
            '--from-offset'. If negative, gives an offset from the end of the
            file. Default: the end of the file.
        """,
    )

    range_group.add_argument(
        "--offset",
        "-o",
        metavar="BIT_OFFSET",
        type=int,
        help="""
            Shows the parts of the bitstream surrounding the specified bit
            offset. (An alternative to manually setting '--from-offset' and
            '--to-offset' to values either-side of the chosen offset). If
            none of '--after-context', '--before-context' or '--context' are
            given, {} bits of context either side of this offset will be
            shown.
        """.format(
            DEFAULT_CONTEXT_BITS
        ),
    )

    range_group.add_argument(
        "--after-context",
        "-A",
        metavar="NUM_BITS",
        type=int,
        help="""
            Sets the number of bits after '--around-offset' to be shown.
        """,
    )
    range_group.add_argument(
        "--before-context",
        "-B",
        metavar="NUM_BITS",
        type=int,
        help="""
            Sets the number of bits before '--around-offset' to be shown.
        """,
    )
    range_group.add_argument(
        "--context",
        "-C",
        metavar="NUM_BITS",
        type=int,
        help="""
            Sets the number of bits before and after '--around-offset' to be
            shown.
        """,
    )

    ###########################################################################

    filter_group = parser.add_argument_group(title="filtering options")

    filter_group.add_argument(
        "--show",
        "-s",
        type=str,
        metavar="FUNCTION",
        default=[],
        action="append",
        help="""
            Display only parts of the bitstream which are processed by the
            specified pseudo-code function as described in the VC-2
            specification. By default all parts of the bitstream are displayed.
            Can be used multiple times to show different parts of the
            bitstream. Supported function names: {}.
        """.format(
            ", ".join(sorted(bitstream.pseudocode_function_to_fixeddicts_recursive))
        ),
    )

    filter_group.add_argument(
        "--hide",
        "-H",
        type=str,
        metavar="FUNCTION",
        default=[],
        action="append",
        help="""
            Omit parts of the bitstream which are processed by the specified
            pseudo-code function as described in the VC-2 specification.
            Accepts the same values as --show.
        """,
    )

    filter_group.add_argument(
        "--hide-slice",
        "-S",
        action="append_const",
        const="slice",
        dest="hide",
        help="""
            Alias for '--hide slice'. Suppresses the printing of all transform
            coefficients, greatly reducing the quantity of output.
        """,
    )

    ###########################################################################

    args = parser.parse_args(*args, **kwargs)

    from_to_offset = args.from_offset is not None or args.to_offset is not None
    if from_to_offset and args.offset is not None:
        parser.error("--offset cannot be used with --from-offset or --to-offset")

    # Convert the '--offset' argument into '--from-offset' and '--to-offset'
    # arguments.
    if args.offset is not None:
        # Make sure the --*-context arguments don't conflict
        if args.context is not None and (
            args.after_context is not None or args.before_context is not None
        ):
            parser.error(
                "--context cannot be used at the same time as --after-context or --before-context"
            )

        before_context = after_context = None

        if args.context is not None:
            before_context = after_context = args.context
        if args.before_context is not None:
            before_context = args.before_context
        if args.after_context is not None:
            after_context = args.after_context

        if before_context is None and after_context is None:
            before_context = after_context = DEFAULT_CONTEXT_BITS
        elif before_context is None:
            before_context = 0
        elif after_context is None:
            after_context = 0

        if args.offset < 0:
            args.from_offset = args.offset - before_context
            args.to_offset = (False, min(-1, args.offset + after_context))
        else:
            args.from_offset = max(0, args.offset - before_context)
            args.to_offset = (False, args.offset + after_context)

        args.offset = None
        args.context = None
        args.before_context = None
        args.after_context = None

    # Make sure the --*-context arguments aren't used with --{from,to}-offset
    if args.after_context is not None:
        parser.error("--after-context can only be used with --offset")
    if args.before_context is not None:
        parser.error("--before-context can only be used with --offset")
    if args.context is not None:
        parser.error("--context can only be used with --offset")

    # Set defaults for '--{from,to}-offset'
    if args.from_offset is None:
        args.from_offset = 0
    if args.to_offset is None:
        args.to_offset = (False, -1)

    # Normalise '--to-offset' into just an integer
    relative, to_offset = args.to_offset
    if relative:
        args.to_offset = args.from_offset + to_offset
    else:
        args.to_offset = to_offset

    # Check show/hide only contain allowed values
    for name in args.show:
        if name not in bitstream.pseudocode_function_to_fixeddicts_recursive:
            parser.error(
                "--show includes unrecognised pseudocode function {!r}".format(name)
            )
    for name in args.hide:
        if name not in bitstream.pseudocode_function_to_fixeddicts_recursive:
            parser.error(
                "--hide includes unrecognised pseudocode function {!r}".format(name)
            )

    return args


def main(*args, **kwargs):
    args = parse_args(*args, **kwargs)

    viewer = BitstreamViewer(
        filename=args.bitstream,
        from_offset=args.from_offset,
        to_offset=args.to_offset,
        shown_pseudocode_names=args.show,
        hidden_pseudocode_names=args.hide,
        show_internal_state=args.show_internal_state,
        check_parse_info_prefix=not args.ignore_parse_info_prefix,
        show_status=not args.no_status,
        verbose=args.verbose,
        num_trailing_bits_on_error=args.num_trailing_bits,
    )
    try:
        return viewer.run()
    finally:
        viewer.close()


if __name__ == "__main__":
    sys.exit(main())



"""
A :py:mod:`~vc2_conformance.fixeddict` for the 'state' object used by the
pseudocode in the VC-2 spec.
"""

from vc2_conformance.string_formatters import Hex

from vc2_data_tables import (
    ParseCodes,
    WaveletFilters,
    Profiles,
    Levels,
)

from vc2_conformance.fixeddict import fixeddict, Entry

__all__ = [
    "State",
    "reset_state",
]

State = fixeddict(
    "State",
    # (10.4.1) parse_sequence
    Entry(
        "video_parameters",
        help_type=":py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`",
        help="Set by (10.4.1) parse_sequence",
    ),
    # (10.5.1) parse_info
    Entry(
        "parse_code",
        enum=ParseCodes,
        formatter=Hex(2),
        help_type=":py:class:`~vc2_data_tables.ParseCodes`",
        help="Set by (10.5.1) parse_info",
    ),
    Entry("next_parse_offset", help_type="int", help="Set by (10.5.1) parse_info"),
    Entry("previous_parse_offset", help_type="int", help="Set by (10.5.1) parse_info"),
    # (11.1) sequence_header
    Entry(
        "picture_coding_mode",
        help_type=":py:class:`~vc2_data_tables.PictureCodingModes`",
        help="Set by (11.1) sequence_header",
    ),
    # (11.2.1) parse_parameters
    Entry("major_version", help_type="int", help="Set by (11.2.1) parse_parameters"),
    Entry("minor_version", help_type="int", help="Set by (11.2.1) parse_parameters"),
    Entry(
        "profile",
        enum=Profiles,
        help_type=":py:class:`~vc2_data_tables.Profiles`",
        help="Set by (11.2.1) parse_parameters",
    ),
    Entry(
        "level",
        enum=Levels,
        help_type=":py:class:`~vc2_data_tables.Profiles`",
        help="Set by (11.2.1) parse_parameters",
    ),
    # (11.6.2) picture_dimensions
    Entry("luma_width", help_type="int", help="Set by (11.6.2) picture_dimensions"),
    Entry("luma_height", help_type="int", help="Set by (11.6.2) picture_dimensions"),
    Entry(
        "color_diff_width", help_type="int", help="Set by (11.6.2) picture_dimensions"
    ),
    Entry(
        "color_diff_height", help_type="int", help="Set by (11.6.2) picture_dimensions"
    ),
    # (11.6.3) video_depth
    Entry("luma_depth", help_type="int", help="Set by (11.6.3) video_depth"),
    Entry("color_diff_depth", help_type="int", help="Set by (11.6.3) video_depth"),
    # (12.2) picture_header and (14.2) fragment_header
    Entry(
        "picture_number",
        help_type="int",
        help="Set by (12.2) picture_header and (14.2) fragment_header",
    ),
    # (12.4.1) transform_parameters
    Entry(
        "wavelet_index",
        enum=WaveletFilters,
        help_type=":py:class:`~vc2_data_tables.WaveletFilters`",
        help="Set by (12.4.1) transform_parameters",
    ),
    Entry("dwt_depth", help_type="int", help="Set by (12.4.1) transform_parameters"),
    # (12.4.4.1) extended_transform_parameters
    Entry(
        "wavelet_index_ho",
        enum=WaveletFilters,
        help_type=":py:class:`~vc2_data_tables.WaveletFilters`",
        help="Set by (12.4.4.1) extended_transform_parameters",
    ),
    Entry("dwt_depth_ho"),
    Entry(
        "dwt_depth_ho",
        help_type="int",
        help="Set by (12.4.4.1) extended_transform_parameters",
    ),
    # (12.4.5.2) slice_parameters
    Entry("slices_x", help_type="int", help="Set by (12.4.5.2) slice_parameters"),
    Entry("slices_y", help_type="int", help="Set by (12.4.5.2) slice_parameters"),
    Entry(
        "slice_bytes_numerator",
        help_type="int",
        help="Set by (12.4.5.2) slice_parameters",
    ),
    Entry(
        "slice_bytes_denominator",
        help_type="int",
        help="Set by (12.4.5.2) slice_parameters",
    ),
    Entry(
        "slice_prefix_bytes", help_type="int", help="Set by (12.4.5.2) slice_parameters"
    ),
    Entry(
        "slice_size_scaler", help_type="int", help="Set by (12.4.5.2) slice_parameters"
    ),
    # (12.4.5.3) quant_matrix
    Entry(
        "quant_matrix",
        help_type="{level: {orient: int, ...}, ...}",
        help="Set by (12.4.5.3) quant_matrix",
    ),
    # (13.5.5) slice_quantizers
    Entry(
        "quantizer",
        help_type="{level: {orient: int, ...}, ...}",
        help="Set by (13.5.5) slice_quantizers",
    ),
    # (13.5.6.3) slice_band and (13.5.6.4) color_diff_slice_band
    Entry(
        "y_transform",
        help_type="{level: {orient: [[int, ...], ...], ...}, ...}",
        help="""
            Set by (13.5.6.3) slice_band. The dequantised luma transform data
            read from the bitstream. A 2D array for each subband, indexed as
            [level][orient][y][x].
        """,
    ),
    Entry(
        "c1_transform",
        help_type="{level: {orient: [[int, ...], ...], ...}, ...}",
        help="""
            Set by (13.5.6.3) slice_band and (13.5.6.4)
            color_diff_slice_band. The dequantised color difference 1 transform
            data read from the bitstream. A 2D array for each subband, indexed
            as [level][orient][y][x].
        """,
    ),
    Entry(
        "c2_transform",
        help_type="{level: {orient: [[int, ...], ...], ...}, ...}",
        help="""
            Set by (13.5.6.3) slice_band and (13.5.6.4)
            color_diff_slice_band. The dequantised color difference 2 transform
            data read from the bitstream. A 2D array for each subband, indexed
            as [level][orient][y][x].
        """,
    ),
    # (14.2) fragment_header
    Entry(
        "fragment_data_length", help_type="int", help="Set by (14.2) fragment_header"
    ),
    Entry(
        "fragment_slice_count", help_type="int", help="Set by (14.2) fragment_header"
    ),
    Entry("fragment_x_offset", help_type="int", help="Set by (14.2) fragment_header"),
    Entry("fragment_y_offset", help_type="int", help="Set by (14.2) fragment_header"),
    # (14.4) fragment_data
    Entry(
        "fragment_slices_received", help_type="int", help="Set by (14.4) fragment_data"
    ),
    Entry(
        "fragmented_picture_done", help_type="int", help="Set by (14.4) fragment_data"
    ),
    # (15.2) picture_decode
    Entry(
        "current_picture",
        help_type=(
            "{'pic_num': int, "
            "'Y': [[int, ...], ...], "
            "'C1': [[int, ...], ...], "
            "'C2': [[int, ...], ...]}"
        ),
        help="Set by (15.2) picture_decode, contains the decoded picture.",
    ),
    # (A.2.1) read_*
    Entry("next_bit", help_type="int", help="Set by (A.2.1) read_*"),
    Entry("current_byte", help_type="int", help="Set by (A.2.1) read_*"),
    # (A.4.2) bounded blocks
    Entry(
        "bits_left",
        help_type="int",
        help="Bits left in the current bounded block (A.4.2)",
    ),
    # The following values are not part of the VC-2 spec but are used by
    # vc2_conformance.decoder to interface with the outside world (which is
    # beyond the scope of the VC-2 spec).
    #
    # (10.4.1) parse_sequence related state
    Entry(
        "_generic_sequence_matcher",
        help_type=":py:class:`vc2_conformance.symbol_re.Matcher`",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            A :py:class:`~vc2_conformance.symbol_re.Matcher` which checks that
            the sequence follows the specified general pattern of data units
            (10.4.1) (e.g. start with a sequence header, end with end of
            sequence).  Other matchers will test for, e.g. level-defined
            patterns.
        """,
    ),
    # (15.2) output_picture related state
    Entry(
        "_output_picture_callback",
        help_type=(
            "function("
            "{'pic_num': int, "
            "'Y': [[int, ...], ...], "
            "'C1': [[int, ...], ...], "
            "'C2': [[int, ...], ...]}, "
            ":py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`, "
            ":py:class:`~vc2_data_tables.PictureCodingModes`"
            ")"
        ),
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            A callback function to call when ``output_picture`` (15.2) is
            called. This callback (if defined) will be passed the picture,
            video parameters and picture coding mode.
        """,
    ),
    # (10.4.3) and (12.2)
    Entry(
        "_num_pictures_in_sequence",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (10.4.3) and (12.2) A counter of how many pictures have been
            encountered in the current sequence. Used to determine if all
            fields have been received (when pictures are fields) and that the
            earliest fields have an even picture number. Initialised in
            parse_sequence (10.4.1) and incremented in
            :py:func:`vc2_conformance.decoder.assertions.assert_picture_number_incremented_as_expected`,
            called in
            picture_header (12.2), and fragment_header (14.2).
        """,
    ),
    # (10.5.1) parse_info related state
    Entry(
        "_last_parse_info_offset",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (10.5.1) parse_info related state. The byte offset of the
            previously read parse_info block.
        """,
    ),
    # (11.1) sequence_header
    Entry(
        "_recorded_bytes",
        help_type=":py:class:`bytearray`",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (11.1) sequence_header requires that repeats must be byte-for-byte
            identical. To facilitate this, the
            :py:func:`vc2_conformance.decoder.io.record_bitstream_start`
            and :py:func:`vc2_conformance.decoder.io.record_bitstream_finish`
            functions are used to make a recording.  When this entry is absent,
            read_byte works as usual. If it is a :py:class:`bytearray`,
            whenever a byte has been completely read, it will be placed into
            this array.
        """,
    ),
    Entry(
        "_last_sequence_header_bytes",
        help_type=":py:class:`bytearray`",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (11.1) the bitstream bytes of the data which encoded the previous
            sequence_header in the sequence. Not present if no previous
            sequence_header has appeared.
        """,
    ),
    Entry(
        "_last_sequence_header_offset",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (11.1) the bitstream offset in bytes of the data which encoded the
            previous sequence_header in the sequence. Not present if no
            previous sequence_header has appeared.
        """,
    ),
    # (11.2.2) Version number constraint checking
    Entry(
        "_expected_major_version",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`. Used to
            record the expected major_version (11.2.1) for this sequence
            according to the constraints listed in (11.2.2). This field is set
            by
            :py:func:`~vc2_conformance.decoder.assertions.log_version_lower_bound`.
        """,
    ),
    # (12.2) picture_header and (14.2) fragment_header
    Entry(
        "_last_picture_number_offset",
        help_type="(byte offset, next bit)",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (12.2) picture_header and (14.2) fragment_header: The offset of the
            last picture number encountered in the sequence (or absent if no
            pictures have yet been encountered).
        """,
    ),
    Entry(
        "_last_picture_number",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (12.2) picture_header and (14.2) fragment_header: The last picture
            number encountered in the sequence (or absent if no pictures have
            yet been encountered).
        """,
    ),
    # (14) fragments
    Entry(
        "_picture_initial_fragment_offset",
        help_type="(byte offset, next bit)",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (14) The offset in the bitstream of the last fragment with
            fragment_slice_count==0.
        """,
    ),
    Entry(
        "_fragment_slices_remaining",
        help_type="int",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            (14) The number of fragment slices which remain to be received in
            the current picture. Initialised to zero by parse_sequence
            (10.4.1), reset to the number of slices per picture by
            initialize_fragment_state (14.3) and decremented whenever a slice
            is received by fragment_data (14.4).
        """,
    ),
    # (A.2.1) read_byte
    Entry(
        "_file",
        help_type="file-like",
        help="""
            The Python file-like object from which the bitstream will be read
            by read_byte (A.2.1) .
        """,
    ),
    # (C.3) Level-related state
    Entry(
        "_level_sequence_matcher",
        help_type=":py:class:`vc2_conformance.symbol_re.Matcher`",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            A :py:class:`~vc2_conformance.symbol_re.Matcher` which checks that
            the sequence follows the pattern dictated by the current level.
            Populated after the first (11.2.1) parse_parameters is encountered
            (and should therefore be ignored until that point).
        """,
    ),
    Entry(
        "_level_constrained_values",
        help_type="{key: value, ...}",
        help="""
            Not in spec, used by :py:mod:`vc2_conformance.decoder`.
            A dictionary which will be incrementally populated with values read or
            computed from the bitstream which are constrained by the level
            constraints table (see
            :py:data:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS` and
            :py:func:`vc2_conformance.decoder.assertions.assert_level_constraint`).
        """,
    ),
    help="""
        The global state variable type.

        Entries prefixed with an underscore (``_``) are not specified by the
        VC-2 pseudocode but may be used by the conformance software.
    """,
)


retained_state_fields = [
    # The output_picture callback should remain so that subsequent sequences
    # trigger the same callback.
    "_output_picture_callback",
    # I/O state must be preserved to allow continuing to read the current file
    "next_bit",
    "current_byte",
    "_file",
    "_recorded_bytes",  # Not required in practice, here for consistency
]
"""
The list of fields in the :py:class:`State` dictionary which should be retained
between sequences in a stream.
"""


def reset_state(state):
    """
    Reset a :py:class:`~vc2_conformance.pseudocode.state.State` dictionary to only include
    values retained between sequences in a VC-2 stream. Modifies the dictionary
    in place.
    """
    for key in set(state.keys()) - set(retained_state_fields):
        del state[key]



"""
Shared utilities used by several test case generators.
"""

from io import BytesIO

from vc2_data_tables import (
    Profiles,
    ParseCodes,
    PARSE_INFO_HEADER_BYTES,
)

from vc2_conformance.pseudocode.state import State

from vc2_conformance.bitstream import (
    Serialiser,
    BitstreamWriter,
    ParseInfo,
    parse_info,
    vc2_default_values,
)


def make_dummy_end_of_sequence(previous_parse_offset=PARSE_INFO_HEADER_BYTES):
    """
    Make (and serialise) an end-of-sequence data unit to be placed within a
    padding data unit.
    """
    f = BytesIO()
    state = State()
    context = ParseInfo(
        parse_code=ParseCodes.end_of_sequence,
        next_parse_offset=0,
        previous_parse_offset=previous_parse_offset,
    )
    with Serialiser(BitstreamWriter(f), context, vc2_default_values) as ser:
        parse_info(ser, state)

    return f.getvalue()


def iter_transform_parameters_in_sequence(codec_features, sequence):
    """
    Iterate over all of the transform parameters in a sequence.

    Generates a series of
    :py:class:`~vc2_conformance.bitstream.TransformParameters` dicts, one for
    each picture present in the provided
    :py:class:`~vc2_conformance.bitstream.Sequence`.

    NB: This function assumes the stream is conformant.
    """
    for data_unit in sequence["data_units"]:
        # Get the TransformParameters for the current picture/fragment data
        # unit
        if "picture_parse" in data_unit:
            tp = data_unit["picture_parse"]["wavelet_transform"]["transform_parameters"]
            yield tp
        elif "fragment_parse" in data_unit:
            if "transform_parameters" in data_unit["fragment_parse"]:
                yield data_unit["fragment_parse"]["transform_parameters"]


def iter_slices_in_sequence(codec_features, sequence):
    """
    Iterate over all of the slices in a sequence.

    Generates a series of (:py:class:`~vc2_conformance.pseudocode.state.State`, sx, sy,
    :py:class:`~vc2_conformance.bitstream.LDSlice` or
    :py:class:`~vc2_conformance.bitstream.HQSlice`) tuples, one for each slice
    present in the provided :py:class:`~vc2_conformance.bitstream.Sequence`.
    The state dictionary will be populated as required by the two
    ``fill_*_slice_padding`` functions.

    NB: This function assumes the stream is conformant (i.e. has the correct
    number of slices, all fragments are present and in order etc).
    """
    sx = sy = 0

    state = None

    for data_unit in sequence["data_units"]:
        # Get the TransformData/FragmentData (td_or_fd) and SliceParameters
        # (sp) for the current picture/fragment data unit
        td_or_fd = None
        sp = None
        if "picture_parse" in data_unit:
            wt = data_unit["picture_parse"]["wavelet_transform"]
            td_or_fd = wt["transform_data"]
            sp = wt["transform_parameters"]["slice_parameters"]
        elif "fragment_parse" in data_unit:
            if "fragment_data" in data_unit["fragment_parse"]:
                td_or_fd = data_unit["fragment_parse"]["fragment_data"]
            if "transform_parameters" in data_unit["fragment_parse"]:
                sp = data_unit["fragment_parse"]["transform_parameters"][
                    "slice_parameters"
                ]

        # Got transform parameters
        if sp is not None:
            if codec_features["profile"] == Profiles.high_quality:
                state = State(
                    slice_prefix_bytes=sp["slice_prefix_bytes"],
                    slice_size_scaler=sp["slice_size_scaler"],
                    slices_x=sp["slices_x"],
                    slices_y=sp["slices_y"],
                )
            elif codec_features["profile"] == Profiles.low_delay:
                state = State(
                    slice_bytes_numerator=sp["slice_bytes_numerator"],
                    slice_bytes_denominator=sp["slice_bytes_denominator"],
                    slices_x=sp["slices_x"],
                    slices_y=sp["slices_y"],
                )

        # Got some slices
        if td_or_fd is not None:
            if codec_features["profile"] == Profiles.high_quality:
                slices = td_or_fd["hq_slices"]
            elif codec_features["profile"] == Profiles.low_delay:
                slices = td_or_fd["ld_slices"]

            for slice in slices:
                yield (state, sx, sy, slice)

                sx += 1
                if sx >= state["slices_x"]:
                    sx = 0

                    sy += 1
                    if sy >= state["slices_y"]:
                        sy = 0



"""
The :py:mod:`vc2_conformance.bitstream.io` module contains low-level wrappers
for file-like objects which facilitate bitwise read and write operations of the
kinds used by VC-2's bitstream format.

The :py:class:`BitstreamReader` and :py:class:`BitstreamWriter` classes provide
equivalent methods for the various ``read_*`` pseudocode functions defined in
the VC-2 specification, along with a few additional utility methods.

.. note::

    These methods are designed to be 'safe' meaning that if out-of-range values
    are provided an error will be produced (rather than an unexpected value
    being written/read).

.. autoclass:: BitstreamReader
    :members:

.. autoclass:: BitstreamWriter
    :members:

The following utility functions are also provided for converting between
offsets given as ``(bytes, bits)`` pairs and offsets given in bytes.

.. autofunction:: to_bit_offset

.. autofunction:: from_bit_offset

"""

from bitarray import bitarray

from vc2_conformance.string_formatters import Bytes

from vc2_conformance.bitstream.exceptions import OutOfRangeError


__all__ = [
    "to_bit_offset",
    "from_bit_offset",
    "BitstreamReader",
    "BitstreamWriter",
]


def to_bit_offset(bytes, bits=7):
    """
    Convert from a (bytes, bits) tuple (as used by
    :py:meth:`BitstreamReader.tell` and :py:meth:`BitstreamWriter.tell`) into a
    total number of bits.
    """
    return (bytes * 8) + (7 - bits)


def from_bit_offset(total_bits):
    """
    Convert from a bit offset into a (bytes, bits) tuple (as used by
    :py:meth:`BitstreamReader.tell` and :py:meth:`BitstreamWriter.tell`).
    """
    bytes = total_bits // 8
    bits = total_bits % 8

    return (bytes, 7 - bits)


class BitstreamReader(object):
    """
    An open file which may be read one bit at a time.

    When the end-of-file is encountered, reads will result in a
    :py:exc:`EOFError`.
    """

    def __init__(self, file):
        """
        Parameters
        ==========
        file : A Python 'file' object in binary-read mode.
        """
        self._file = file

        # The index of the next bit to read
        self._next_bit = 7

        # The current byte index being read
        self._byte_offset = self._file.tell()

        # The byte currently being read (or None if at the EOF)
        self._current_byte = None

        # None, if not in a bounded block. Otherwise, the number of unused bits
        # remaining in the block. If negative, indicates the number of bits
        # read past the end of the block.
        self._bits_remaining = None

        # Load-in the first byte
        self._read_byte()

    def _read_byte(self):
        """Internal method. Advance to the next byte. (A.2.2)"""
        byte = self._file.read(1)
        self._byte_offset += 1
        if len(byte) == 1:
            self._current_byte = bytearray(byte)[0]
            self._next_bit = 7
        else:
            self._current_byte = None
            self._next_bit = 7

    def is_end_of_stream(self):
        """Check if we've reached the EOF. (A.2.5)"""
        return self._current_byte is None

    def tell(self):
        """
        Report the current bit-position within the stream.

        Returns
        =======
        (bytes, bits)
            ``bytes`` is the offset of the current byte from the start of the
            stream. ``bits`` is the offset in the current byte (starting at 7
            (MSB) and advancing towards 0 (LSB) as bits are read).
        """
        return (self._byte_offset - 1, self._next_bit)

    def seek(self, bytes, bits=7):
        """
        Seek to a specific (absolute) position in the file.

        Parameters
        ==========
        bytes : int
            The byte-offset from the start of the file.
        bits : int
            The bit offset into the specified byte to start reading from.
        """
        assert 0 <= bits <= 7

        # Special case: in a bounded block
        if self._bits_remaining is not None:
            new_offset = to_bit_offset(bytes, bits)
            cur_offset = to_bit_offset(*self.tell())
            delta = new_offset - cur_offset

            # Special case: If the seek attemts to move past the end of the
            # bounded block, this is not possible so fail
            if delta > 0 and self._bits_remaining - delta < 0:
                raise Exception("Cannot seek() past end of bounded block.")

            if self._bits_remaining <= 0 and delta == 0:
                # We're past the end but not moving, don't change the count
                pass
            elif self._bits_remaining < 0 and delta < 0:
                # We're currently past the end of the bounded block and the
                # seek moves us back before the block, reset the bits remaining
                # accordingly
                self._bits_remaining = -delta
            else:
                # We're moving, adjust the remaining bit count accordingly
                self._bits_remaining -= delta

        self._file.seek(bytes)
        self._byte_offset = self._file.tell()
        self._read_byte()
        self._next_bit = bits

    @property
    def bits_remaining(self):
        """
        The number of bits left in the current bounded block.

        None, if not in a bounded block. Otherwise, the number of unused bits
        remaining in the block. If negative, indicates the number of bits read
        past the end of the block.
        """
        return self._bits_remaining

    def bounded_block_begin(self, length):
        """
        Begin a bounded block of the specified length in bits.
        """
        if self._bits_remaining is not None:
            raise Exception("Cannot nest bounded blocks")
        self._bits_remaining = length

    def bounded_block_end(self):
        """
        Ends the current bounded block. Returns the number of unused bits
        remaining, but does not read them or seek past them.
        """
        if self._bits_remaining is None:
            raise Exception("Not in bounded block.")

        unused_bits = max(0, self._bits_remaining)
        self._bits_remaining = None

        return unused_bits

    def read_bit(self):
        """
        Read and return the next bit in the stream. (A.2.3) Reads '1' for
        values past the end of file.
        """
        # Don't read past the end of the current bounded block
        if self._bits_remaining is not None:
            self._bits_remaining -= 1

            # NB: Checked *after* decrement hence <= -1 not <= 0
            if self._bits_remaining <= -1:
                return 1

        if self._current_byte is None:
            raise EOFError()

        bit = (self._current_byte >> self._next_bit) & 1

        self._next_bit -= 1
        if self._next_bit < 0:
            self._read_byte()

        return bit

    def read_nbits(self, bits):
        """
        Read an 'bits'-bit unsigned integer (like read_nbits (A.3.3)).
        """
        value = 0
        for i in range(bits):
            value <<= 1
            value |= self.read_bit()

        return value

    def read_uint_lit(self, num_bytes):
        """
        Read a 'num-bytes' long integer (like read_uint_lit (A.3.4)).
        """
        return self.read_nbits(num_bytes * 8)

    def read_bitarray(self, bits):
        """
        Read 'bits' bits returning the value as a
        :py:class:`bitarray.bitarray`.
        """
        return bitarray(self.read_bit() for _ in range(bits))

    def read_bytes(self, num_bytes):
        """
        Read a number of bytes returning a :py:class:`bytes` string.
        """
        return self.read_bitarray(num_bytes * 8).tobytes()

    def read_uint(self):
        """
        Read an unsigned exp-golomb code (like read_uint (A.4.3)) and return an
        integer.
        """
        value = 1
        while True:
            if self.read_bit():
                break
            else:
                value <<= 1
                value += self.read_bit()

        value -= 1

        return value

    def read_sint(self):
        """
        Signed version of :py:meth:`read_uint` (like read_sint (A.4.4)).
        """
        value = self.read_uint()

        # Read sign bit
        if value != 0:
            if self.read_bit():
                value = -value

        return value

    def try_read_bitarray(self, bits):
        """
        Attempt to read the next 'bits' bits from the bitstream file, leaving
        any bounded blocks we might be in if necessary). May read fewer bits if
        the end-of-file is encountered (but will not throw a :py:exc:`EOFError`
        like other methods of this class).

        Intended for the display of error messages (i.e. as the final use of a
        :py:class:`BitstreamReader` instance) only since this method may (or
        may not) exit the current bounded block as a side effect.
        """
        out = bitarray()

        try:
            for _ in range(bits):
                if self.bits_remaining is not None and self.bits_remaining <= 0:
                    self.bounded_block_end()
                out.append(self.read_bit())
        except EOFError:
            pass

        return out


class BitstreamWriter(object):
    """
    An open file which may be written one bit at a time.
    """

    def __init__(self, file):
        """
        Parameters
        ==========
        file : A Python 'file' object in binary-write mode.
        """
        self._file = file

        # The index of the next bit to write
        self._next_bit = 7

        # The current byte index being written
        self._byte_offset = self._file.tell()

        # The byte currently being write
        self._current_byte = 0

        # None, if not in a bounded block. Otherwise, the number of unused bits
        # remaining in the block. If negative, indicates the number of bits
        # read past the end of the block.
        self._bits_remaining = None

    def _write_byte(self):
        """Internal method. Write the current byte and start a new one. (A.2.2)"""
        self._file.write(bytearray([self._current_byte]))
        self._current_byte = 0
        self._next_bit = 7
        self._byte_offset += 1

    def is_end_of_stream(self):
        """
        Always True. (A.2.5)

        .. note::

            Strictly speaking this should return False when seeking to an
            earlier part of the stream however this behaviour is not
            implemented here for simplicity's sake.
        """
        return True

    def tell(self):
        """
        Report the current bit-position within the stream.

        Returns
        =======
        (bytes, bits)
            ``bytes`` is the offset of the current byte from the start of the
            stream. ``bits`` is the offset in the current byte (starting at 7
            (MSB) and advancing towards 0 (LSB) as bits are written).
        """
        return (self._byte_offset, self._next_bit)

    def seek(self, bytes, bits=7):
        """
        Seek to a specific (absolute) position in the file. Seeking to a given
        byte will overwrite any bits already set in that byte to 0.

        Parameters
        ==========
        bytes : int
            The byte-offset from the start of the file.
        bits : int
            The bit offset into the specified byte to start writing to.
        """
        assert 0 <= bits <= 7

        # Special case: in a bounded block
        if self._bits_remaining is not None:
            new_offset = to_bit_offset(bytes, bits)
            cur_offset = to_bit_offset(*self.tell())
            delta = new_offset - cur_offset

            # Special case: If the seek attemts to move past the end of the
            # bounded block, this is not possible so fail
            if delta > 0 and self._bits_remaining - delta < 0:
                raise Exception("Cannot seek() past end of bounded block.")

            if self._bits_remaining <= 0 and delta == 0:
                # We're past the end but not moving, don't change the count
                pass
            elif self._bits_remaining < 0 and delta < 0:
                # We're currently past the end of the bounded block and the
                # seek moves us back before the block, reset the bits remaining
                # accordingly
                self._bits_remaining = -delta
            else:
                # We're moving, adjust the remaining bit count accordingly
                self._bits_remaining -= delta

        self.flush()

        self._file.seek(bytes)
        self._byte_offset = self._file.tell()
        self._current_byte = 0
        self._next_bit = bits

    def flush(self):
        """
        Ensure all bytes are committed to the file.
        """
        # Write the current byte to the file (if any bits have been written
        # into it). Note that we don't use write_bit() since this write is
        # writing a partially complete byte and so we don't wish to move on yet
        # (the user may later write some more bits to this byte).
        if self._next_bit != 7:
            self._file.write(bytearray([self._current_byte]))
            self._file.seek(-1, 1)  # Seek backward 1 byte

        self._file.flush()

    @property
    def bits_remaining(self):
        """
        The number of bits left in the current bounded block.

        None, if not in a bounded block. Otherwise, the number of unused bits
        remaining in the block. If negative, indicates the number of bits read
        past the end of the block.
        """
        return self._bits_remaining

    def bounded_block_begin(self, length):
        """
        Begin a bounded block of the specified length in bits.
        """
        if self._bits_remaining is not None:
            raise Exception("Cannot nest bounded blocks")
        self._bits_remaining = length

    def bounded_block_end(self):
        """
        Ends the current bounded block. Returns the number of unused bits
        remaining, but does not write them or seek past them.
        """
        if self._bits_remaining is None:
            raise Exception("Not in bounded block.")

        unused_bits = max(0, self._bits_remaining)
        self._bits_remaining = None

        return unused_bits

    def write_bit(self, value):
        """
        Write a bit into the bitstream. If in a bounded block, raises a
        :py:exc:`ValueError` if a '0' is written beyond the end of the block.
        """
        # Don't write past the end of the current bounded block
        if self._bits_remaining is not None:
            self._bits_remaining -= 1

            # NB: Checked *after* decrement hence <= -1 not <= 0
            if self._bits_remaining <= -1:
                if not value:
                    raise ValueError("Cannot write 0s past the end of a bounded block.")
                return

        # Clear the target bit
        self._current_byte &= ~(1 << self._next_bit)
        # Set new value
        if value:
            self._current_byte |= 1 << self._next_bit

        self._next_bit -= 1
        if self._next_bit < 0:
            self._write_byte()

    def write_nbits(self, bits, value):
        """
        Write an 'bits'-bit integer. The complement of read_nbits (A.3.3).

        Throws an :py:exc:`OutOfRangeError` if the value is too large to fit in
        the requested number of bits.
        """
        if value < 0 or value.bit_length() > bits:
            raise OutOfRangeError(
                "0b{:b} is {} bits, not {}".format(
                    value,
                    value.bit_length(),
                    bits,
                )
            )

        for i in range(bits - 1, -1, -1):
            self.write_bit((value >> i) & 1)

    def write_uint_lit(self, num_bytes, value):
        """
        Write a 'num-bytes' long integer. The complement of read_uint_lit (A.3.4).

        Throws an :py:exc:`OutOfRangeError` if the value is too large to fit in
        the requested number of bytes.
        """
        self.write_nbits(num_bytes * 8, value)

    def write_bitarray(self, bits, value):
        """
        Write the 'bits' from the :py;class:`bitarray.bitarray` 'value'.

        Throws an :py:exc:`OutOfRangeError` if the value is longer than 'bits'.
        The value will be right-hand zero-padded to the required length.
        """
        if len(value) > bits:
            raise OutOfRangeError(
                "0b{} is {} bits, not {}".format(
                    value.to01(),
                    len(value),
                    bits,
                )
            )

        for bit in value:
            self.write_bit(bit)

        # Zero-pad
        for _ in range(len(value), bits):
            self.write_bit(0)

    def write_bytes(self, num_bytes, value):
        """
        Write the provided :py:class:`bytes` or :py:class:`bytearray` in a python
        bytestring.

        If the provided byte string is too long an :py:exc:`OutOfRangeError`
        will be raised. If it is too short, it will be right-hand zero-padded.
        """
        if len(value) > num_bytes:
            raise OutOfRangeError(
                "{} is {} bytes, not {}".format(
                    Bytes()(value),
                    len(value),
                    num_bytes,
                )
            )

        for byte in bytearray(value):
            self.write_nbits(8, byte)

        # Zero-pad
        for _ in range(len(value), num_bytes):
            self.write_nbits(8, 0)

    def write_uint(self, value):
        """
        Write an unsigned exp-golomb code.

        An :py:exc:`OutOfRangeError` will be raised if a negative value is
        provided.
        """
        if value < 0:
            raise OutOfRangeError("{} is negative, expected positive".format(value))

        value += 1

        for i in range(value.bit_length() - 2, -1, -1):
            self.write_bit(0)
            self.write_bit((value >> i) & 1)

        self.write_bit(1)

    def write_sint(self, value):
        """
        Signed version of :py:meth:`write_uint`.
        """
        self.write_uint(abs(value))

        # Write sign bit
        if value != 0:
            self.write_bit(value < 0)



r"""
The :py:mod:`vc2_conformance.bitstream.vc2` module contains a set of
:py:class:`.SerDes`-using (see :py:mod:`~vc2_conformance.bitstream.serdes`)
functions which follow the pseudo-code in the VC-2 specification as closely as
possible. All pseudocode functions are re-exported by the
:py:mod:`vc2_conformance.bitstream` module.

See the table in :ref:`bitstream-fixeddicts` which relates these functions to
their matching :py:mod:`fixeddicts <vc2_conformance.fixeddict>`.

In this module, all functions are derived from the pseudocode by:

* Replacing ``read_*`` calls with :py:class:`.SerDes` calls.
* Adding :py:class:`.SerDes` annotations.
* Removing of decoding functionality (retaining only the code required for
  bitstream deserialisation).

Consistency with the VC-2 pseudocode is checked by the test suite (see
:py:mod:`verification`).
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.bitstream.serdes import context_type

from vc2_data_tables import (
    BaseVideoFormats,
    PresetFrameRates,
    PresetPixelAspectRatios,
    PresetSignalRanges,
    PresetColorSpecs,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    PARSE_INFO_HEADER_BYTES,
)

from vc2_conformance.pseudocode.state import reset_state

from vc2_conformance.pseudocode.video_parameters import (
    set_source_defaults,
    set_coding_parameters,
    preset_frame_rate,
    preset_pixel_aspect_ratio,
    preset_signal_range,
    preset_color_primaries,
    preset_color_matrix,
    preset_transfer_function,
    preset_color_spec,
)

from vc2_conformance.pseudocode.parse_code_functions import (
    is_seq_header,
    is_end_of_sequence,
    is_auxiliary_data,
    is_padding_data,
    is_picture,
    is_ld,
    is_hq,
    is_fragment,
)

from vc2_conformance.pseudocode.vc2_math import intlog2

from vc2_conformance.pseudocode.slice_sizes import (
    slice_bytes,
    slice_left,
    slice_right,
    slice_top,
    slice_bottom,
)

from vc2_conformance.bitstream.vc2_fixeddicts import (
    ParseInfo,
    SequenceHeader,
    ParseParameters,
    SourceParameters,
    FrameSize,
    ColorDiffSamplingFormat,
    ScanFormat,
    FrameRate,
    PixelAspectRatio,
    CleanArea,
    SignalRange,
    ColorSpec,
    ColorPrimaries,
    ColorMatrix,
    TransferFunction,
    AuxiliaryData,
    Padding,
    TransformParameters,
    ExtendedTransformParameters,
    SliceParameters,
    QuantMatrix,
    PictureHeader,
    TransformData,
    WaveletTransform,
    PictureParse,
    LDSlice,
    HQSlice,
    FragmentHeader,
    FragmentData,
    FragmentParse,
    DataUnit,
    Sequence,
    Stream,
)

__all__ = [
    "parse_stream",
    "parse_sequence",
    "auxiliary_data",
    "padding",
    "parse_info",
    "sequence_header",
    "parse_parameters",
    "source_parameters",
    "frame_size",
    "color_diff_sampling_format",
    "scan_format",
    "frame_rate",
    "pixel_aspect_ratio",
    "clean_area",
    "signal_range",
    "color_spec",
    "color_primaries",
    "color_matrix",
    "transfer_function",
    "picture_parse",
    "picture_header",
    "wavelet_transform",
    "transform_data",
    "slice",
    "ld_slice",
    "hq_slice",
    "slice_band",
    "color_diff_slice_band",
    "transform_parameters",
    "extended_transform_parameters",
    "slice_parameters",
    "quant_matrix",
    "fragment_parse",
    "fragment_header",
    "fragment_data",
]


################################################################################
# (10) Stream decoding
################################################################################


@context_type(Stream)
@ref_pseudocode(deviation="serdes")
def parse_stream(serdes, state):
    """
    (10.3) Parse a VC-2 whole stream containing multiple concatenated
    sequences, discarding the contents.

    Populates a :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.Stream`
    :py:mod:`~vc2_conformance.fixeddict`.
    """
    serdes.declare_list("sequences")

    # Here the logic for deciding when a sequence is 'over' is complicated by
    # the fact that the pseudocode uses an EOF signal to determine this. In the
    # Serialisation case, this obviously isn't suitable so we instead wait for
    # the 'sequences' list to run out of entries.
    #
    ### while not is_end_of_stream(state):
    ## Begin not in spec
    while (
        # Deserialisation case: stream is ended by an end-of-stream (NB
        # BitstreamWriter.is_end_of_stream always returns True)
        not serdes.io.is_end_of_stream()
        # Serialisation case: stream is ended by running out of sequences to
        # serialise...
        or not serdes.is_target_complete("sequences")
    ):
        ## End not in spec
        with serdes.subcontext("sequences"):
            parse_sequence(serdes, state)


@context_type(Sequence)
@ref_pseudocode(deviation="serdes")
def parse_sequence(serdes, state):
    """
    (10.4.1) Parse a VC-2 sequence, discarding the contents.

    Populates a :py:class:`~vc2_conformance.bitstream.vc2_fixeddicts.Sequence`
    :py:mod:`~vc2_conformance.fixeddict`.  Provides a copy of the
    :py:class:`~vc2_conformance.pseudocode.state.State` in ``"_state"``.
    """
    reset_state(state)

    serdes.computed_value("_state", state)

    serdes.declare_list("data_units")

    serdes.subcontext_enter("data_units")
    serdes.set_context_type(DataUnit)
    with serdes.subcontext("parse_info"):
        parse_info(serdes, state)
    while not is_end_of_sequence(state):
        if is_seq_header(state):
            with serdes.subcontext("sequence_header"):
                state["video_parameters"] = sequence_header(serdes, state)
        elif is_picture(state):
            with serdes.subcontext("picture_parse"):
                picture_parse(serdes, state)
            # No need to perform the IDWT
            ### picture_decode(state)
        elif is_fragment(state):
            with serdes.subcontext("fragment_parse"):
                fragment_parse(serdes, state)
            # No need to perform the IDWT
            ### if state["fragmented_picture_done"]:
            ###     picture_decode(state)
        elif is_auxiliary_data(state):
            with serdes.subcontext("auxiliary_data"):
                auxiliary_data(serdes, state)
        elif is_padding_data(state):
            with serdes.subcontext("padding"):
                padding(serdes, state)
        serdes.subcontext_leave()

        serdes.subcontext_enter("data_units")
        serdes.set_context_type(DataUnit)
        with serdes.subcontext("parse_info"):
            parse_info(serdes, state)

    serdes.subcontext_leave()


@context_type(ParseInfo)
@ref_pseudocode(deviation="serdes")
def parse_info(serdes, state):
    """(10.5.1) Read a parse_info header."""
    serdes.byte_align("padding")

    ## Begin not in spec
    serdes.computed_value("_offset", serdes.io.tell()[0])
    ## End not in spec

    serdes.uint_lit("parse_info_prefix", 4)
    state["parse_code"] = serdes.uint_lit("parse_code", 1)
    state["next_parse_offset"] = serdes.uint_lit("next_parse_offset", 4)
    state["previous_parse_offset"] = serdes.uint_lit("previous_parse_offset", 4)


@context_type(AuxiliaryData)
@ref_pseudocode(deviation="serdes")
def auxiliary_data(serdes, state):
    """(10.4.4) Read an auxiliary data block."""
    ### for i in range(1, state["next_parse_offset"]-12):
    ###     read_uint_lit(state, 1)
    ## Begin not in spec
    serdes.bytes("bytes", state["next_parse_offset"] - PARSE_INFO_HEADER_BYTES)
    ## End not in spec


@context_type(Padding)
@ref_pseudocode(deviation="serdes")
def padding(serdes, state):
    """(10.4.5) Read a padding data block."""
    ### for i in range(1, state["next_parse_offset"]-12):
    ###     read_uint_lit(state, 1)
    ## Begin not in spec
    serdes.bytes("bytes", state["next_parse_offset"] - PARSE_INFO_HEADER_BYTES)
    ## End not in spec


################################################################################
# (11) Sequence header parsing
################################################################################


@context_type(SequenceHeader)
@ref_pseudocode(deviation="serdes")
def sequence_header(serdes, state):
    """(11.1) Parse a sequence header returning a VideoParameters object."""
    with serdes.subcontext("parse_parameters"):
        parse_parameters(serdes, state)

    base_video_format = serdes.uint("base_video_format")

    # For robustness against bad bitstreams, force unrecognised base formats to 'custom'
    ## Begin not in spec
    try:
        BaseVideoFormats(base_video_format)
    except ValueError:
        base_video_format = BaseVideoFormats.custom_format
    ## End not in spec

    with serdes.subcontext("video_parameters"):
        video_parameters = source_parameters(serdes, state, base_video_format)
    state["picture_coding_mode"] = serdes.uint("picture_coding_mode")
    set_coding_parameters(state, video_parameters)

    return video_parameters


@context_type(ParseParameters)
@ref_pseudocode(deviation="serdes")
def parse_parameters(serdes, state):
    """(11.2.1)"""
    state["major_version"] = serdes.uint("major_version")
    state["minor_version"] = serdes.uint("minor_version")
    state["profile"] = serdes.uint("profile")
    state["level"] = serdes.uint("level")


@context_type(SourceParameters)
@ref_pseudocode(deviation="serdes")
def source_parameters(serdes, state, base_video_format):
    """
    (11.4.1) Parse the video source parameters. Returns a VideoParameters
    object.
    """
    video_parameters = set_source_defaults(base_video_format)

    with serdes.subcontext("frame_size"):
        frame_size(serdes, state, video_parameters)
    with serdes.subcontext("color_diff_sampling_format"):
        color_diff_sampling_format(serdes, state, video_parameters)
    with serdes.subcontext("scan_format"):
        scan_format(serdes, state, video_parameters)
    with serdes.subcontext("frame_rate"):
        frame_rate(serdes, state, video_parameters)
    with serdes.subcontext("pixel_aspect_ratio"):
        pixel_aspect_ratio(serdes, state, video_parameters)
    with serdes.subcontext("clean_area"):
        clean_area(serdes, state, video_parameters)
    with serdes.subcontext("signal_range"):
        signal_range(serdes, state, video_parameters)
    with serdes.subcontext("color_spec"):
        color_spec(serdes, state, video_parameters)

    return video_parameters


@context_type(FrameSize)
@ref_pseudocode(deviation="serdes")
def frame_size(serdes, state, video_parameters):
    """(11.4.3) Override video parameter."""
    custom_dimensions_flag = serdes.bool("custom_dimensions_flag")
    if custom_dimensions_flag:
        video_parameters["frame_width"] = serdes.uint("frame_width")
        video_parameters["frame_height"] = serdes.uint("frame_height")


@context_type(ColorDiffSamplingFormat)
@ref_pseudocode(deviation="serdes")
def color_diff_sampling_format(serdes, state, video_parameters):
    """(11.4.4) Override color sampling parameter."""
    custom_color_diff_format_flag = serdes.bool("custom_color_diff_format_flag")
    if custom_color_diff_format_flag:
        video_parameters["color_diff_format_index"] = serdes.uint(
            "color_diff_format_index"
        )


@context_type(ScanFormat)
@ref_pseudocode(deviation="serdes")
def scan_format(serdes, state, video_parameters):
    """(11.4.5) Override source sampling parameter."""
    custom_scan_format_flag = serdes.bool("custom_scan_format_flag")
    if custom_scan_format_flag:
        video_parameters["source_sampling"] = serdes.uint("source_sampling")


@context_type(FrameRate)
@ref_pseudocode(deviation="serdes")
def frame_rate(serdes, state, video_parameters):
    """(11.4.6) Override frame-rate parameter."""
    custom_frame_rate_flag = serdes.bool("custom_frame_rate_flag")

    if custom_frame_rate_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            if index != 0:
                PresetFrameRates(index)
        except ValueError:
            index = PresetFrameRates.fps_24_over_1_001
        ## End not in spec

        if index == 0:
            video_parameters["frame_rate_numer"] = serdes.uint("frame_rate_numer")
            video_parameters["frame_rate_denom"] = serdes.uint("frame_rate_denom")
        else:
            preset_frame_rate(video_parameters, index)


@context_type(PixelAspectRatio)
@ref_pseudocode(deviation="serdes")
def pixel_aspect_ratio(serdes, state, video_parameters):
    """(11.4.7) Override pixel aspect ratio parameter."""
    custom_pixel_aspect_ratio_flag = serdes.bool("custom_pixel_aspect_ratio_flag")
    if custom_pixel_aspect_ratio_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            if index != 0:
                PresetPixelAspectRatios(index)
        except ValueError:
            index = PresetPixelAspectRatios.ratio_1_1
        ## End not in spec

        if index == 0:
            video_parameters["pixel_aspect_ratio_numer"] = serdes.uint(
                "pixel_aspect_ratio_numer"
            )
            video_parameters["pixel_aspect_ratio_denom"] = serdes.uint(
                "pixel_aspect_ratio_denom"
            )
        else:
            preset_pixel_aspect_ratio(video_parameters, index)


@context_type(CleanArea)
@ref_pseudocode(deviation="serdes")
def clean_area(serdes, state, video_parameters):
    """(11.4.8) Override clean area parameter."""
    custom_clean_area_flag = serdes.bool("custom_clean_area_flag")
    if custom_clean_area_flag:
        video_parameters["clean_width"] = serdes.uint("clean_width")
        video_parameters["clean_height"] = serdes.uint("clean_height")
        video_parameters["left_offset"] = serdes.uint("left_offset")
        video_parameters["top_offset"] = serdes.uint("top_offset")


@context_type(SignalRange)
@ref_pseudocode(deviation="serdes")
def signal_range(serdes, state, video_parameters):
    """(11.4.9) Override signal parameter."""
    custom_signal_range_flag = serdes.bool("custom_signal_range_flag")
    if custom_signal_range_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            if index != 0:
                PresetSignalRanges(index)
        except ValueError:
            index = PresetSignalRanges.video_8bit_full_range
        ## End not in spec

        if index == 0:
            video_parameters["luma_offset"] = serdes.uint("luma_offset")
            video_parameters["luma_excursion"] = serdes.uint("luma_excursion")
            video_parameters["color_diff_offset"] = serdes.uint("color_diff_offset")
            video_parameters["color_diff_excursion"] = serdes.uint(
                "color_diff_excursion"
            )
        else:
            preset_signal_range(video_parameters, index)


@context_type(ColorSpec)
@ref_pseudocode(deviation="serdes")
def color_spec(serdes, state, video_parameters):
    """(11.4.10.1) Override color specification parameter."""
    custom_color_spec_flag = serdes.bool("custom_color_spec_flag")
    if custom_color_spec_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            PresetColorSpecs(index)
        except ValueError:
            index = PresetColorSpecs.sdtv_525
        ## End not in spec

        preset_color_spec(video_parameters, index)
        if index == 0:
            with serdes.subcontext("color_primaries"):
                color_primaries(serdes, state, video_parameters)
            with serdes.subcontext("color_matrix"):
                color_matrix(serdes, state, video_parameters)
            with serdes.subcontext("transfer_function"):
                transfer_function(serdes, state, video_parameters)


@context_type(ColorPrimaries)
@ref_pseudocode(deviation="serdes")
def color_primaries(serdes, state, video_parameters):
    """(11.4.10.2) Override color primaries parameter."""
    custom_color_primaries_flag = serdes.bool("custom_color_primaries_flag")
    if custom_color_primaries_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            PresetColorPrimaries(index)
        except ValueError:
            index = PresetColorPrimaries.hdtv
        ## End not in spec

        preset_color_primaries(video_parameters, index)


@context_type(ColorMatrix)
@ref_pseudocode(deviation="serdes")
def color_matrix(serdes, state, video_parameters):
    """(11.4.10.3) Override color matrix parameter."""
    custom_color_matrix_flag = serdes.bool("custom_color_matrix_flag")
    if custom_color_matrix_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            PresetColorMatrices(index)
        except ValueError:
            index = PresetColorMatrices.hdtv
        ## End not in spec

        preset_color_matrix(video_parameters, index)


@context_type(TransferFunction)
@ref_pseudocode(deviation="serdes")
def transfer_function(serdes, state, video_parameters):
    """(11.4.10.4) Override color transfer function parameter."""
    custom_transfer_function_flag = serdes.bool("custom_transfer_function_flag")
    if custom_transfer_function_flag:
        index = serdes.uint("index")

        # Required to prevent crashes on malformed inputs make an arbitrary
        # choice
        ## Begin not in spec
        try:
            PresetTransferFunctions(index)
        except ValueError:
            index = PresetTransferFunctions.tv_gamma
        ## End not in spec

        preset_transfer_function(video_parameters, index)


################################################################################
# (12) Picture syntax
################################################################################


@context_type(PictureParse)
@ref_pseudocode(deviation="serdes")
def picture_parse(serdes, state):
    """(12.1)"""
    serdes.byte_align("padding1")
    with serdes.subcontext("picture_header"):
        picture_header(serdes, state)
    serdes.byte_align("padding2")
    with serdes.subcontext("wavelet_transform"):
        wavelet_transform(serdes, state)


@context_type(PictureHeader)
@ref_pseudocode(deviation="serdes")
def picture_header(serdes, state):
    """(12.2)"""
    state["picture_number"] = serdes.uint_lit("picture_number", 4)


@context_type(WaveletTransform)
@ref_pseudocode(deviation="serdes")
def wavelet_transform(serdes, state):
    """(12.3)"""
    with serdes.subcontext("transform_parameters"):
        transform_parameters(serdes, state)
    serdes.byte_align("padding")
    with serdes.subcontext("transform_data"):
        transform_data(serdes, state)


@context_type(TransformParameters)
@ref_pseudocode(deviation="serdes")
def transform_parameters(serdes, state):
    """(12.4.1) Read transform parameters."""
    state["wavelet_index"] = serdes.uint("wavelet_index")
    state["dwt_depth"] = serdes.uint("dwt_depth")

    state["wavelet_index_ho"] = state["wavelet_index"]
    state["dwt_depth_ho"] = 0
    if state["major_version"] >= 3:
        with serdes.subcontext("extended_transform_parameters"):
            extended_transform_parameters(serdes, state)

    with serdes.subcontext("slice_parameters"):
        slice_parameters(serdes, state)
    with serdes.subcontext("quant_matrix"):
        quant_matrix(serdes, state)


@context_type(ExtendedTransformParameters)
@ref_pseudocode(deviation="serdes")
def extended_transform_parameters(serdes, state):
    """(12.4.4.1) Read horizontal only transform parameters."""
    asym_transform_index_flag = serdes.bool("asym_transform_index_flag")
    if asym_transform_index_flag:
        state["wavelet_index_ho"] = serdes.uint("wavelet_index_ho")

    asym_transform_flag = serdes.bool("asym_transform_flag")
    if asym_transform_flag:
        state["dwt_depth_ho"] = serdes.uint("dwt_depth_ho")


@context_type(SliceParameters)
@ref_pseudocode(deviation="serdes")
def slice_parameters(serdes, state):
    """(12.4.5.2) Read slice parameters"""
    state["slices_x"] = serdes.uint("slices_x")
    state["slices_y"] = serdes.uint("slices_y")

    if is_ld(state):
        state["slice_bytes_numerator"] = serdes.uint("slice_bytes_numerator")
        state["slice_bytes_denominator"] = serdes.uint("slice_bytes_denominator")
    if is_hq(state):
        state["slice_prefix_bytes"] = serdes.uint("slice_prefix_bytes")
        state["slice_size_scaler"] = serdes.uint("slice_size_scaler")


@context_type(QuantMatrix)
@ref_pseudocode(deviation="serdes")
def quant_matrix(serdes, state):
    """(12.4.5.3) Read quantisation matrix"""
    custom_quant_matrix = serdes.bool("custom_quant_matrix")
    if custom_quant_matrix:
        serdes.declare_list("quant_matrix")

        # NB: For historical reasons, we use a dict not an array in this
        # implementation.
        ### state["quant_matrix"] = new_array(state["dwt_depth_ho"] + state["dwt_depth"] + 1)
        state["quant_matrix"] = {}  ## Not in spec
        if state["dwt_depth_ho"] == 0:
            state["quant_matrix"][0] = {}
            state["quant_matrix"][0]["LL"] = serdes.uint("quant_matrix")
        else:
            state["quant_matrix"][0] = {}
            state["quant_matrix"][0]["L"] = serdes.uint("quant_matrix")
            for level in range(1, state["dwt_depth_ho"] + 1):
                state["quant_matrix"][level] = {}
                state["quant_matrix"][level]["H"] = serdes.uint("quant_matrix")

        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            state["quant_matrix"][level] = {}
            state["quant_matrix"][level]["HL"] = serdes.uint("quant_matrix")
            state["quant_matrix"][level]["LH"] = serdes.uint("quant_matrix")
            state["quant_matrix"][level]["HH"] = serdes.uint("quant_matrix")
    else:
        # Not required for bitstream decoding
        ### set_quant_matrix(state)
        pass  ## Not in spec


################################################################################
# (13) Transform data syntax
################################################################################


@context_type(TransformData)
@ref_pseudocode(deviation="serdes")
def transform_data(serdes, state):
    """(13.5.2)"""
    # Not required for bitstream decoding
    ### state["y_transform"] = initialize_wavelet_data(state, "Y")
    ### state["c1_transform"] = initialize_wavelet_data(state, "C1")
    ### state["c2_transform"] = initialize_wavelet_data(state, "C2")

    # Initialise context dictionary
    ## Begin not in spec
    serdes.computed_value("_state", state.copy())
    if is_ld(state):
        serdes.declare_list("ld_slices")
    if is_hq(state):
        serdes.declare_list("hq_slices")
    ## End not in spec

    for sy in range(state["slices_y"]):
        for sx in range(state["slices_x"]):
            slice(serdes, state, sx, sy)

    # Not required for bitstream decoding
    ### if using_dc_prediction(state):
    ###     if state["dwt_depth_ho"] == 0:
    ###         dc_prediction(state["y_transform"][0]["LL"])
    ###         dc_prediction(state["c1_transform"][0]["LL"])
    ###         dc_prediction(state["c2_transform"][0]["LL"])
    ###     else:
    ###         dc_prediction(state["y_transform"][0]["L"])
    ###         dc_prediction(state["c1_transform"][0]["L"])
    ###         dc_prediction(state["c2_transform"][0]["L"])

    # Following line included to ensure the above comment is considered part of
    # this function
    pass  ## Not in spec


@ref_pseudocode(deviation="serdes")
def slice(serdes, state, sx, sy):
    """(13.5.2)"""
    if is_ld(state):
        with serdes.subcontext("ld_slices"):
            ld_slice(serdes, state, sx, sy)
    elif is_hq(state):
        with serdes.subcontext("hq_slices"):
            hq_slice(serdes, state, sx, sy)


@context_type(LDSlice)
@ref_pseudocode(deviation="serdes")
def ld_slice(serdes, state, sx, sy):
    """(13.5.3.1)"""
    slice_bits_left = 8 * slice_bytes(state, sx, sy)

    qindex = serdes.nbits("qindex", 7)  # noqa: F841
    slice_bits_left -= 7

    # Not required for bitstream unpacking
    ### slice_quantizers(state, qindex)

    length_bits = intlog2((8 * slice_bytes(state, sx, sy)) - 7)
    slice_y_length = serdes.nbits("slice_y_length", length_bits)
    slice_bits_left -= length_bits

    # Ensure robustness in the presence of invalid bitstreams
    ## Begin not in spec
    if slice_y_length > slice_bits_left:
        slice_y_length = slice_bits_left
    ## End not in spec

    # Initialise context dictionary
    serdes.computed_value("_sx", sx)
    serdes.computed_value("_sy", sy)
    serdes.declare_list("y_transform")
    serdes.declare_list("c_transform")

    serdes.bounded_block_begin(slice_y_length)

    if state["dwt_depth_ho"] == 0:
        slice_band(serdes, state, "y_transform", 0, "LL", sx, sy)
        for level in range(1, state["dwt_depth"] + 1):
            for orient in ["HL", "LH", "HH"]:
                slice_band(serdes, state, "y_transform", level, orient, sx, sy)
    else:
        slice_band(serdes, state, "y_transform", 0, "L", sx, sy)
        for level in range(1, state["dwt_depth_ho"] + 1):
            slice_band(serdes, state, "y_transform", level, "H", sx, sy)
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            for orient in ["HL", "LH", "HH"]:
                slice_band(serdes, state, "y_transform", level, orient, sx, sy)

    serdes.bounded_block_end("y_block_padding")

    slice_bits_left -= slice_y_length

    serdes.bounded_block_begin(slice_bits_left)

    if state["dwt_depth_ho"] == 0:
        color_diff_slice_band(serdes, state, 0, "LL", sx, sy)
        for level in range(1, state["dwt_depth"] + 1):
            for orient in ["HL", "LH", "HH"]:
                color_diff_slice_band(serdes, state, level, orient, sx, sy)
    else:
        color_diff_slice_band(serdes, state, 0, "L", sx, sy)
        for level in range(1, state["dwt_depth_ho"] + 1):
            color_diff_slice_band(serdes, state, level, "H", sx, sy)
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            for orient in ["HL", "LH", "HH"]:
                color_diff_slice_band(serdes, state, level, orient, sx, sy)

    serdes.bounded_block_end("c_block_padding")


@context_type(HQSlice)
@ref_pseudocode(deviation="serdes")
def hq_slice(serdes, state, sx, sy):
    """(13.5.4)"""
    serdes.bytes("prefix_bytes", state["slice_prefix_bytes"])

    qindex = serdes.uint_lit("qindex", 1)  # noqa: F841

    # Not required for bitstream unpacking
    ### slice_quantizers(state, qindex)

    # Initialise context dictionary
    serdes.computed_value("_sx", sx)
    serdes.computed_value("_sy", sy)
    serdes.declare_list("y_transform")
    serdes.declare_list("c1_transform")
    serdes.declare_list("c2_transform")

    for transform in ["y_transform", "c1_transform", "c2_transform"]:
        component = transform.split("_")[0]  ## Not in spec
        length = state["slice_size_scaler"] * serdes.uint_lit(
            "slice_{}_length".format(component), 1
        )

        serdes.bounded_block_begin(8 * length)

        if state["dwt_depth_ho"] == 0:
            slice_band(serdes, state, transform, 0, "LL", sx, sy)
            for level in range(1, state["dwt_depth"] + 1):
                for orient in ["HL", "LH", "HH"]:
                    slice_band(serdes, state, transform, level, orient, sx, sy)
        else:
            slice_band(serdes, state, transform, 0, "L", sx, sy)
            for level in range(1, state["dwt_depth_ho"] + 1):
                slice_band(serdes, state, transform, level, "H", sx, sy)
            for level in range(
                state["dwt_depth_ho"] + 1,
                state["dwt_depth_ho"] + state["dwt_depth"] + 1,
            ):
                for orient in ["HL", "LH", "HH"]:
                    slice_band(serdes, state, transform, level, orient, sx, sy)

        serdes.bounded_block_end("{}_block_padding".format(component))


@ref_pseudocode(deviation="serdes")
def slice_band(serdes, state, transform, level, orient, sx, sy):
    """(13.5.6.3) Read and dequantize a subband in a slice."""
    if transform == "y_transform":
        comp = "Y"
    elif transform == "c1_transform":
        comp = "C1"
    elif transform == "c2_transform":
        comp = "C2"

    # Not required for bitstream unpacking
    ### qi = state["quantizer"][level][orient]

    # These values evaulated in the loop definition in the spec, moving them
    # here saves a lot of computation
    ## Begin not in spec
    y1 = slice_top(state, sy, comp, level)
    y2 = slice_bottom(state, sy, comp, level)
    x1 = slice_left(state, sx, comp, level)
    x2 = slice_right(state, sx, comp, level)
    ## End not in spec

    ### for y in range(slice_top(state, sy,comp,level), slice_bottom(state, sy,comp,level)):
    ###     for x in range(slice_left(state, sx,comp,level), slice_right(state, sx,comp,level)):
    for y in range(y1, y2):  ## Not in spec
        for x in range(x1, x2):  ## Not in spec
            val = serdes.sint(transform)  # noqa: F841

            # Not required for bitstream unpacking
            ### state[transform][level][orient][y][x] = inverse_quant(val, qi)

    # Following line included to ensure the trailing comment above is
    # considered part of this function
    pass  ## Not in spec


@ref_pseudocode(deviation="serdes")
def color_diff_slice_band(serdes, state, level, orient, sx, sy):
    """(13.5.6.4) Read and dequantize interleaved color difference subbands in a slice."""
    # Not required for bitstream unpacking
    ### qi = state["quantizer"][level][orient]

    # These values evaulated in the loop definition in the spec, moving them
    # here saves a lot of computation
    ## Begin not in spec
    y1 = slice_top(state, sy, "C1", level)
    y2 = slice_bottom(state, sy, "C1", level)
    x1 = slice_left(state, sx, "C1", level)
    x2 = slice_right(state, sx, "C1", level)
    ## End not in spec

    ### for y in range(slice_top(state,sy,"C1",level), slice_bottom(state,sy,"C1",level)):
    ###     for x in range(slice_left(state,sx,"C1",level), slice_right(state,sx,"C1",level)):
    for y in range(y1, y2):  ## Not in spec
        for x in range(x1, x2):  ## Not in spec
            val = serdes.sint("c_transform")  # noqa: F841
            ### state["c1_transform"][level][orient][y][x] = inverse_quant(val, qi)

            val = serdes.sint("c_transform")  # noqa: F841
            ### state["c2_transform"][level][orient][y][x] = inverse_quant(val, qi)

    # Following line included to ensure the trailing comment above is
    # considered part of this function
    pass  ## Not in spec


################################################################################
# (14) Fragment syntax
################################################################################


@context_type(FragmentParse)
@ref_pseudocode(deviation="serdes")
def fragment_parse(serdes, state):
    """(14.1)"""
    with serdes.subcontext("fragment_header"):
        fragment_header(serdes, state)
    if state["fragment_slice_count"] == 0:
        with serdes.subcontext("transform_parameters"):
            transform_parameters(serdes, state)

        # Not required (prepares structures used for decoding and termination
        # detection)
        ### initialize_fragment_state(state)
    else:
        with serdes.subcontext("fragment_data"):
            fragment_data(serdes, state)


@context_type(FragmentHeader)
@ref_pseudocode(deviation="serdes")
def fragment_header(serdes, state):
    """(14.2)"""
    state["picture_number"] = serdes.uint_lit("picture_number", 4)
    state["fragment_data_length"] = serdes.uint_lit("fragment_data_length", 2)
    state["fragment_slice_count"] = serdes.uint_lit("fragment_slice_count", 2)
    if state["fragment_slice_count"] != 0:
        state["fragment_x_offset"] = serdes.uint_lit("fragment_x_offset", 2)
        state["fragment_y_offset"] = serdes.uint_lit("fragment_y_offset", 2)


@context_type(FragmentData)
@ref_pseudocode(deviation="serdes")
def fragment_data(serdes, state):
    """(14.4) Unpack and dequantize transform data from a fragment."""
    # Initialise context dictionary
    ## Begin not in spec
    serdes.computed_value("_state", state.copy())
    if is_ld(state):
        serdes.declare_list("ld_slices")
    if is_hq(state):
        serdes.declare_list("hq_slices")
    ## End not in spec

    for s in range(state["fragment_slice_count"]):
        slice_x = (
            (state["fragment_y_offset"] * state["slices_x"])
            + state["fragment_x_offset"]
            + s
        ) % state["slices_x"]
        slice_y = (
            (state["fragment_y_offset"] * state["slices_x"])
            + state["fragment_x_offset"]
            + s
        ) // state["slices_x"]
        slice(serdes, state, slice_x, slice_y)

        # Not required for bitstream unpacking
        ### state["fragment_slices_received"] += 1
        ###
        ### if state["fragment_slices_received"] == (state["slices_x"] * state["slices_y"]):
        ###     state["fragmented_picture_done"] = True
        ###     if using_dc_prediction(state):
        ###         if state["dwt_depth_ho"] == 0:
        ###             dc_prediction(state["y_transform"][0]["LL"])
        ###             dc_prediction(state["c1_transform"][0]["LL"])
        ###             dc_prediction(state["c2_transform"][0]["LL"])
        ###         else:
        ###             dc_prediction(state["y_transform"][0]["L"])
        ###             dc_prediction(state["c1_transform"][0]["L"])
        ###             dc_prediction(state["c2_transform"][0]["L"])

    # Following line included to ensure the trailing comment above is
    # considered part of this function
    pass  ## Not in spec
